import pytest
import os,sys,inspect
import numpy as np
import pyscal.core as pc
from pyscal.crystal_structures import Structure


def test_rdf_bcc():
    sys = Structure().lattice.bcc(repetitions = [10, 10, 10])
    rdf, r = sys.calculate_rdf(rmax=2)

    args = np.argsort(rdf)[::-1]
    assert(r[args[0]]-0.86 < 1E-5)

def test_rdf_fcc():
    sys = Structure().lattice.fcc(repetitions = [10, 10, 10])
    rdf, r = sys.calculate_rdf(rmax=2)

    args = np.argsort(rdf)[::-1]
    assert(r[args[0]]-0.70 < 1E-5)



import pytest
import os
import numpy as np
import pyscal.core as pc
from pyscal.crystal_structures import Structure

def test_q_4():
    sys = Structure().lattice.bcc(repetitions = [4, 4, 4])

    #sys.get_neighbors(method = 'voronoi')
    sys.find_neighbors(method = 'cutoff', cutoff=0.9)

    q = sys.calculate_q([4, 6])
    assert np.round(np.mean(np.array(q[0])), decimals=2) == 0.51 , "Calculated q4 value is wrong!"
    assert np.round(np.mean(np.array(q[1])), decimals=2) == 0.63 , "Calculated q4 value is wrong!"

    q = sys.calculate_q([4, 6], averaged=True)
    assert np.round(np.mean(np.array(q[0])), decimals=2) == 0.51 , "Calculated q4 value is wrong!"
    assert np.round(np.mean(np.array(q[1])), decimals=2) == 0.63 , "Calculated q4 value is wrong!"




"""
This file includes the definition of an Atoms class which can be used with
System

TODO
----
- Iterators for masked/unmasked atoms
- Iterators for selected/unselected atoms
"""

import numpy as np
import warnings
import os
from pyscal.attributes import AttrSetter, read_yaml, MyList

attr_docs = read_yaml(os.path.join(os.path.dirname(__file__), "data/annotations.yaml"))

class Atoms(dict, AttrSetter):
    def __init__(self, atoms=None):
        #self.update(atoms=atoms)
        self._nreal = 0
        self._nghost = 0
        self._lattice_constant = None
        self._lattice = None
        AttrSetter.__init__(self)

        if atoms is not None:
            self.from_dict(atoms)
    
    def __dir__(self):
        attrs = ["natoms", "nreal", "nghost", 
        "ntotal", "from_dict", "iter_atoms", "apply_mask", "remove_mask",
        "apply_selection", "remove_selection", "delete", "composition"]
        return attrs + list(self._map_dict.keys())

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self._get_atoms(key)
        elif isinstance(key, int):
            return self._get_atoms(key)
        else:
            val = dict.__getitem__(self, key)
            return val

    def __setitem__(self, key, val):
        dict.__setitem__(self, key, MyList(val))

    def __repr__(self):
        dictrepr = dict.__repr__(self)
        return '%s(%s)' % (type(self).__name__, dictrepr)

    def _repr_json(self):
        #convert to atom base dict
        disp_atoms = {f"atom {x}": self._get_atoms(x) for x in range(self.natoms)}
        return disp_atoms
        
    #def update(self, atoms):
    #    for k, v in dict(*args, **kwargs).items():
    #        self[k] = v

    def __add__(self, atoms):
        if not 'positions' in atoms.keys():
            raise ValueError('positions is a necessary key in atoms')
        nop = len(atoms["positions"])
        val_length_check = np.prod([len(val)==nop for key, val in atoms.items()])
        if not val_length_check:
            raise ValueError("All times in the atoms dict should have same length as positions")
        
        #now add necessary keys-ids, types, ghost
        maxid = max(self["ids"])
        if not 'ids' in atoms.keys():
            atoms['ids'] = [maxid+x+1 for x in range(nop)]
            #print(self["ids"], atoms["ids"])
        else:
            if len(set(atoms['ids']).intersection(set(self['ids']))):
                raise ValueError("Atom id already exists, unique ID is required")
        
        atoms['ghost'] = [False for x in range(nop)]
        if not 'types' in atoms.keys():
            atoms['types'] = [1 for x in range(nop)]
        if not 'species' in atoms.keys():
            atoms['species'] = [None for x in range(nop)]
        if not 'mask_1' in atoms.keys():
            atoms['mask_1'] = [False for x in range(nop)]
        if not 'mask_2' in atoms.keys():
            atoms['mask_2'] = [False for x in range(nop)]
        if not 'condition' in atoms.keys():
            atoms['condition'] = [True for x in range(nop)]
        if not 'head' in atoms.keys():
            atoms['head'] = [self.natoms+x for x in range(nop)]
        
        common_keys = list(set(self.keys()).intersection(set(atoms.keys())))
        _ = [self[key].extend(atoms[key]) for key in common_keys]
        extra_keys_add = len(atoms.keys()) - len(common_keys)
        extra_keys_exist = len(self.keys()) - len(common_keys)
        
        if extra_keys_add > 0:
            warnings.warn("Some keys present in the atoms are add are not present in existing atoms, they were ignored")
        if extra_keys_exist > 0:
            warnings.warn("Some keys present in the existing Atoms were not present in the atoms to add, please recalculate")
        self._nreal += nop
        return self


    def __radd__(self, atoms):
        """
        Reverse add method
        """
        if ntraj == 0:
            return self
        else:
            return self.__add__(atoms)

    #def _add_attribute()
      
    @property
    def natoms(self):
        return self._nreal

    @property
    def nreal(self):
        return self._nreal
    
    @property
    def nghost(self):
        return self._nghost
    
    @property
    def ntotal(self):
        return self._nreal + self._nghost

    def from_dict(self, atoms):
        if not 'positions' in atoms.keys():
            raise ValueError('positions is a necessary key in atoms')
        nop = len(atoms["positions"])
        
        if not 'ids' in atoms.keys():
            atoms['ids'] = [x+1 for x in range(nop)]
        
        atoms['ghost'] = [False for x in range(nop)]
        if not 'types' in atoms.keys():
            atoms['types'] = [1 for x in range(nop)]
        if not 'species' in atoms.keys():
            atoms['species'] = [None for x in range(nop)]
        if not 'mask_1' in atoms.keys():
            atoms['mask_1'] = [False for x in range(nop)]
        if not 'mask_2' in atoms.keys():
            atoms['mask_2'] = [False for x in range(nop)]
        if not 'condition' in atoms.keys():
            atoms['condition'] = [True for x in range(nop)]
        if not 'head' in atoms.keys():
            atoms['head'] = [x for x in range(nop)]
        
        for key, val in atoms.items():
            self[key] = MyList(val)
        self._nreal = len(val)

        #add attributes
        mapdict = {"positions": "positions",
        "ids": "ids",
        "types": "types",
        "species": "species",
        "mask": {"primary": "mask_1", "secondary": "mask_2"},
        "selection": "condition",
        "condition": "condition",
        "head": "head"}

        #add extra keys that might be needed; non-standard ones
        for key, val in atoms.items():
            if key not in ["positions", "ids", "types", "species", "mask_1", "mask_2", "condition", "head"]:
                mapdict[key] = key

        self._add_attribute(mapdict)

    def _convert_to_list(self, data):
        """
        Check if the given item is a list, if not convert to a single item list
        """
        if not isinstance(data, list):
            data = [data]
        return data
       
    def _get_atoms(self, index):
        atom_dict = {key: self._convert_to_list(self[key][index]) for key in self.keys()}
        return Atoms(atom_dict)

    def _delete_atoms(self, indices):
        del_real = np.sum([1 for x in indices if x < self._nreal])
        del_ghost = np.sum([1 for x in indices if x >= self._nreal])

        for key in self.keys():
            for index in sorted(indices, reverse=True):
                del self[key][index]

        td = len(indices)
        self._nreal = int(self.nreal - del_real)
        self._nghost = int(self.nghost - del_ghost)


    def iter_atoms(self):
        for index in range(self.nreal):
            atom_dict = {key: [self[key][index]] for key in self.keys()}
            yield Atoms(atom_dict)

    def _generate_bool_list(self, ids=None, indices=None, condition=None, selection=False):
        #necessary checks
        non_nones = sum(x is not None for x in [ids, indices, condition])
        if non_nones > 1:
            raise ValueError("Only one of ids, indices or condition should be provided")
        elif ((non_nones == 0) and (selection==False)):
            warnings.warn("No conditions provided, all atoms will be included")
        #generate a list of indices
        if selection:
            indices = [x for x in range(self.nreal) if self["condition"][x]]
        elif ids is not None:
            if not isinstance(ids, list):
                ids = [ids]
            indices = [x for x in range(len(self["ids"])) if self["ids"][x] in ids]
            if len(indices) == 0:
                raise ValueError("No ids found to delete")
            if len(indices) != len(ids):
                warnings.warn("Not all ids were found")
        elif condition is not None:
            indices = [x for x in range(self.nreal) if condition(self._get_atoms(x))]
        elif indices is None:
            indices = [x for x in range(self.nreal)]
        
        if not isinstance(indices, list):
            indices = [indices]

        bool_list = [ True if x in indices else False for x in range(self.nreal)]
        return bool_list

    def _apply_mask(self, masks, mask_type):
        if (mask_type == 'primary') or (mask_type == 'all'):
            for i in range(self.ntotal):
                self["mask_1"][i] = masks[self["head"][i]]
        if (mask_type == 'secondary') or (mask_type == 'all'):
            for i in range(self.ntotal):
                self["mask_2"][i] = masks[self["head"][i]]

    def apply_mask(self, mask_type="primary", ids=None, indices=None, condition=None, selection=False):
        masks = self._generate_bool_list(ids=ids, indices=indices, condition=condition, selection=selection)
        self._apply_mask(masks, mask_type)


    def remove_mask(self, mask_type="primary", ids=None, indices=None, condition=None, selection=False):
        masks = self._generate_bool_list(ids=ids, indices=indices, condition=condition, selection=selection)
        masks = [not x for x in masks]
        self._apply_mask(masks, mask_type)

    def _apply_selection(self, condition):
        for i in range(self.ntotal):
            self["condition"][i] = condition[self["head"][i]]
    
    def _validate_condition(self, condition):
        if not (len(condition)==self.nreal):
            raise ValueError("condition should have same length as atoms")
        for c, x in enumerate(condition):
            try:
                x = bool(x)
                condition[c] = x
            except:
                pass 
            if not isinstance(x, bool):
                raise ValueError("Condition elements should be boolean")
        return condition

    def apply_selection(self, ids=None, indices=None, condition=None):
        if isinstance(condition, list):
            masks = self._validate_condition(condition)
        else:
            masks = self._generate_bool_list(ids=ids, indices=indices, condition=condition)
        self._apply_selection(masks)
    
    def remove_selection(self, ids=None, indices=None, condition=None):
        if isinstance(condition, list):
            masks = self._validate_condition(condition)
        else:
            masks = self._generate_bool_list(ids=ids, indices=indices, condition=condition)
        masks = [not x for x in masks]
        self._apply_selection(masks)

    def delete(self, ids=None, indices=None, condition=None, selection=False):
        #delete atoms
        #reassign ids
        #reassign indices
        #reassign heads
        masks = self._generate_bool_list(ids=ids, indices=indices, condition=condition, selection=selection)
        delete_list = [masks[self["head"][x]] for x in range(self.ntotal)]
        delete_ids = [x for x in range(self.ntotal) if masks[x]]
        self._delete_atoms(delete_ids)
    
    @property
    def _type_dict(self):
        sp = []
        types, typecounts = np.unique(self["types"][:self.nreal], return_counts=True)
        for t in types:
            for count, tx in enumerate(self["types"][:self.nreal]):
                if t==tx:
                    sp.append(self["species"][count])
                    break
        return dict([x for x in zip(types, sp)])


    @property
    def composition(self):
        if self["species"][0] is None:
            typelist = self["types"][:self.nreal]
            types, typecounts = np.unique(typelist, return_counts=True)
            concdict = dict([(t, typecounts[c]/np.sum(typecounts)) for c, t in enumerate(types)])
        else:
            typelist = self["species"][:self.nreal]
            types, typecounts = np.unique(typelist, return_counts=True)
            concdict = {str(t): typecounts[c]/np.sum(typecounts) for c, t in enumerate(types)}
        return concdict



    



import numpy as np
import warnings
import os

from pyscal.atoms import Atoms
from pyscal.attributes import read_yaml
structures = read_yaml(os.path.join(os.path.dirname(__file__), "data/structure_data.yaml"))

def make_crystal(structure, lattice_constant = 1.00, repetitions = None, ca_ratio = 1.633, noise = 0, element=None,
    return_structure_dict=False, structures=structures):
    """
    Create a basic crystal structure and return it as a list of `Atom` objects
    and box dimensions.

    Parameters
    ----------
    structure : {'sc', 'bcc', 'fcc', 'hcp', 'diamond', 'a15' or 'l12'}
        type of the crystal structure

    lattice_constant : float, optional
        lattice constant of the crystal structure, default 1

    repetitions : list of ints of len 3, optional
        of type `[nx, ny, nz]`, repetions of the unit cell in x, y and z directions.
        default `[1, 1, 1]`.

    ca_ratio : float, optional
        ratio of c/a for hcp structures, default 1.633

    noise : float, optional
        If provided add normally distributed noise with standard deviation `noise` to the atomic positions.

    Returns
    -------
    atoms : list of `Atom` objects
        list of all atoms as created by user input

    box : list of list of floats
        list of the type `[[xlow, xhigh], [ylow, yhigh], [zlow, zhigh]]` where each of them are the lower
        and upper limits of the simulation box in x, y and z directions respectively.

    Examples
    --------
    >>> atoms, box = make_crystal('bcc', lattice_constant=3.48, repetitions=[2,2,2])
    >>> sys = System()
    >>> sys.assign_atoms(atoms, box)

    """
    if repetitions == None:
        nx = 1
        ny = 1
        nz = 1
    elif isinstance(repetitions, int):
        nx = repetitions
        ny = repetitions
        nz = repetitions
    else:
        nx = repetitions[0]
        ny = repetitions[1]
        nz = repetitions[2]

    if structure in structures.keys():
        sdict = structures[structure]
    else:
        raise ValueError("Unknown crystal structure")
    
    unique_types = np.unique(sdict["species"])
    print(element)
    if element is not None:
        if isinstance(element, str):
            element = [element]
        if not (len(element) == len(unique_types)):
            raise ValueError("Elements should equal number of species in the system")
        
        element_dict = dict([x for x in zip(unique_types, element)])
    else:
        element = [None for x in range(len(unique_types))]
    element_dict = dict([x for x in zip(unique_types, element)])
        
    m = 0
    co = 1
    natoms = sdict["natoms"]*nx*ny*nz
    positions = []
    types = []
    ids = []
    species = []

    xh = nx*lattice_constant*sdict["scaling_factors"][0]
    yh = ny*lattice_constant*sdict["scaling_factors"][1]
    zh = nz*lattice_constant*sdict["scaling_factors"][2]
    box = [[xh, 0, 0], [0, yh, 0], [0, 0, zh]]

    #create structure
    for i in range(1, nx+1):
        for j in range(1, ny+1):
            for k in range(1, nz+1):
                for l in range(1, sdict["natoms"]+1):
                    m += 1
                    posx = ((lattice_constant*sdict["positions"][l-1][0])+(lattice_constant*sdict["scaling_factors"][0]*(float(i)-1)))
                    posy = ((lattice_constant*sdict["positions"][l-1][1])+(lattice_constant*sdict["scaling_factors"][1]*(float(j)-1)))
                    posz = ((lattice_constant*sdict["positions"][l-1][2])+(lattice_constant*sdict["scaling_factors"][2]*(float(k)-1)))
                    if noise > 0:
                        posx = np.random.normal(loc=posx, scale=noise)
                        posy = np.random.normal(loc=posy, scale=noise)
                        posz = np.random.normal(loc=posz, scale=noise)
                    
                    positions.append([posx, posy, posz])
                    ids.append(co)
                    types.append(sdict["species"][l-1])
                    species.append(element_dict[sdict["species"][l-1]])
                    co += 1
    atoms = {}
    atoms['positions'] = positions
    atoms['ids'] = ids
    atoms['types'] = types
    atoms['species'] = species
    atoms['ghost'] = [False for x in range(len(types))]

    patoms = Atoms()
    patoms.from_dict(atoms)
    if return_structure_dict:
        return patoms, box, sdict

    return patoms, box

def _update_list_of_elements():
    """
    Only run when needed to update database
    """
    import mendeleev
    from mendeleev import element
    el_list = dir(mendeleev)
    el_dict = {}

    for el in el_list:
        if len(el) == 2:
            if el=="db":
                break
            chem = element(el)
            struct = chem.lattice_structure
            lc = chem.lattice_constant
            if (struct is not None) and (lc is not None):
                if struct.lower() in ['sc', 'bcc', 'fcc', 'hex', 'dia']:
                    el_dict[el] = {}
                    if struct.lower() == "dia":
                        struct = "diamond"
                    elif struct.lower() == "hex":
                        struct = "hcp"
                    el_dict[el]["structure"] = str(struct.lower())
                    el_dict[el]["lattice_constant"] = float(lc)
    return el_dict



import pyscal.structure_creator as pcs
import os

def test_create_structure():
    structure = ['bcc', 'fcc', 'hcp', 'diamond', 'a15', 'l12', 'b2']
    natoms = [2, 4, 4, 8, 8, 4, 2]

    for count, struct in enumerate(structure):
        #dumpfile = os.path.join(os.getcwd(), "test.dat")
        atoms, boxdims = pcs.make_crystal(struct)
        assert len(atoms["positions"]) == natoms[count]


        



import sys
import random
from math import degrees, atan, sqrt, pi, ceil, cos, acos, sin, gcd, radians
import numpy as np
from numpy import dot, cross
from numpy.linalg import det, norm, inv

def get_cubic_sigma(uvw, m, n=1):
    """
    CSL analytical formula based on the book:
    'Interfaces in crystalline materials',
     Sutton and Balluffi, clarendon press, 1996.
     generates possible sigma values.
    arguments:
    uvw -- the axis
    m,n -- two integers (n by default 1)

    """
    u, v, w = uvw
    sqsum = u*u + v*v + w*w
    sigma = m*m + n*n * sqsum
    while sigma != 0 and sigma % 2 == 0:
        sigma /= 2
    return sigma if sigma > 1 else None


def get_cubic_theta(uvw, m, n=1):
    """
    generates possible theta values.
    arguments:
    uvw -- the axis
    m,n -- two integers (n by default 1)
    """
    u, v, w = uvw
    sqsum = u*u + v*v + w*w
    if m > 0:
        return 2 * atan(sqrt(sqsum) * n / m)
    else:
        return pi

def get_theta_m_n_list(uvw, sigma):
    """
    Finds integers m and n lists that match the input sigma.
    """
    if sigma == 1:
        return [(0., 0., 0.)]
    thetas = []
    max_m = int(ceil(sqrt(4*sigma)))

    for m in range(1, max_m):
        for n in range(1, max_m):
            if gcd(m, n) == 1:
                s = get_cubic_sigma(uvw, m, n)
            if s == sigma:
                theta = (get_cubic_theta(uvw, m, n))
                thetas.append((theta, m, n))
                thetas.sort(key=lambda x: x[0])
    return thetas

def get_sigma_list(uvw, limit):
    """
    prints a list of smallest sigmas/angles for a given axis(uvw).
    """
    sigmas = []
    thetas = []
    for i in range(limit):
        tt = get_theta_m_n_list(uvw, i)
        if len(tt) > 0:
            theta, _, _ = tt[0]
            sigmas.append(i)
            thetas.append(degrees(theta))
    return sigmas, thetas

def get_theta_m_n_list(uvw, sigma):
    """
    Finds integers m and n lists that match the input sigma.
    """
    if sigma == 1:
        return [(0., 0., 0.)]
    thetas = []
    max_m = int(ceil(sqrt(4*sigma)))

    for m in range(1, max_m):
        for n in range(1, max_m):
            if gcd(m, n) == 1:
                s = get_cubic_sigma(uvw, m, n)
            if s == sigma:
                theta = (get_cubic_theta(uvw, m, n))
                thetas.append((theta, m, n))
                thetas.sort(key=lambda x: x[0])
    return thetas

def rot(a, Theta):
    """
    produces a rotation matrix.
    arguments:
    a -- an axis
    Theta -- an angle
    """
    c = cos(Theta)
    s = sin(Theta)
    a = a / norm(a)
    ax, ay, az = a
    return np.array([[c + ax * ax * (1 - c), ax * ay * (1 - c) - az * s,
                      ax * az * (1 - c) + ay * s],
                    [ay * ax * (1 - c) + az * s, c + ay * ay * (1 - c),
                        ay * az * (1 - c) - ax * s],
                     [az * ax * (1 - c) - ay * s, az * ay * (1 - c) + ax * s,
                      c + az * az * (1 - c)]])


def create_minimal_cell_method_1(sigma, uvw, R):
    """
    finds Minimal cell by means of a numerical search.
    (An alternative analytical method can be used too).
    arguments:
    sigma -- gb sigma
    uvw -- rotation axis
    R -- rotation matrix
    """
    uvw = np.array(uvw)
    MiniCell_1 = np.zeros([3, 3])
    MiniCell_1[:, 2] = uvw

    lim = 20
    x = np.arange(-lim, lim + 1, 1)
    y = x
    z = x
    indice = (np.stack(np.meshgrid(x, y, z)).T).reshape(len(x) ** 3, 3)

    # remove 0 vectors and uvw from the list
    indice_0 = indice[np.where(np.sum(abs(indice), axis=1) != 0)]
    condition1 = ((abs(dot(indice_0, uvw) / norm(indice_0, axis=1) /
                       norm(uvw))).round(7))
    indice_0 = indice_0[np.where(condition1 != 1)]

    if minicell_search(indice_0, MiniCell_1, R, sigma):

        M1, M2 = minicell_search(indice_0, MiniCell_1, R, sigma)
        return (M1, M2)
    else:
        return None


def minicell_search(indices, MiniCell_1, R, sigma):

    tol = 0.001
    # norm1 = norm(indices, axis=1)
    newindices = dot(R, indices.T).T
    nn = indices[np.all(abs(np.round(newindices) - newindices) < 1e-6, axis=1)]
    TestVecs = nn[np.argsort(norm(nn, axis=1))]
    # print(len(indices), len(TestVecs),TestVecs[:20])

    Found = False
    count = 0
    while (not Found) and count < len(TestVecs) - 1:
        if 1 - ang(TestVecs[count], MiniCell_1[:, 2]) > tol:
            # and  (ang(TestVecs[i],uvw) > tol):
            MiniCell_1[:, 1] = (TestVecs[count])
            count += 1
            for j in range(len(TestVecs)):
                if (1 - ang(TestVecs[j], MiniCell_1[:, 2]) > tol) and (
                        1 - ang(TestVecs[j], MiniCell_1[:, 1]) > tol):
                    if (ang(TestVecs[j],
                            cross(MiniCell_1[:, 2], MiniCell_1[:, 1])) > tol):
                        #  The condition that the third vector can not be
                        #  normal to any other two.
                        #  and (ang(TestVecs[i],uvw)> tol) and
                        # (ang(TestVecs[i],MiniCell[:,1])> tol)):
                        MiniCell_1[:, 0] = (TestVecs[j]).astype(int)
                        Det1 = abs(round(det(MiniCell_1), 5))
                        MiniCell_1 = (MiniCell_1).astype(int)
                        MiniCell_2 = ((np.round(dot(R, MiniCell_1), 7))
                                      .astype(int))
                        Det2 = abs(round(det(MiniCell_2), 5))

                        if ((abs(Det1 - sigma)) < tol and
                                (abs(Det2 - sigma)) < tol):
                            Found = True
                            break
    if Found:
        return MiniCell_1, MiniCell_2
    else:
        return Found

def create_possible_gb_plane_list(uvw, m=5, n=1, lim=5):
    """
    generates GB planes and specifies the character.

    arguments:
    uvw -- axis of rotation.
    m,n -- the two necessary integers
    lim -- upper limit for the plane indices

    """
    uvw = np.array(uvw)
    Theta = get_cubic_theta(uvw, m, n)
    Sigma = get_cubic_sigma(uvw, m, n)
    R1 = rot(uvw, Theta)

    # List and character of possible GB planes:
    x = np.arange(-lim, lim + 1, 1)
    y = x
    z = x
    indice = (np.stack(np.meshgrid(x, y, z)).T).reshape(len(x) ** 3, 3)
    indice_0 = indice[np.where(np.sum(abs(indice), axis=1) != 0)]
    indice_0 = indice_0[np.argsort(norm(indice_0, axis=1))]

    # extract the minimal cell:
    Min_1, Min_2 = create_minimal_cell_method_1(Sigma, uvw, R1)
    V1 = np.zeros([len(indice_0), 3])
    V2 = np.zeros([len(indice_0), 3])
    GBtype = []
    tol = 0.001
    # Mirrorplanes cubic symmetry
    MP = np.array([[1, 0, 0],
                   [0, 1, 0],
                   [0, 0, 1],
                   [1, 1, 0],
                   [0, 1, 1],
                   [1, 0, 1],
                   ], dtype='float')
    # Find GB plane coordinates:
    for i in range(len(indice_0)):
        if common_divisor(indice_0[i])[1] <= 1:
            V1[i, :] = (indice_0[i, 0] * Min_1[:, 0] +
                        indice_0[i, 1] * Min_1[:, 1] +
                        indice_0[i, 2] * Min_1[:, 2])
            V2[i, :] = (indice_0[i, 0] * Min_2[:, 0] +
                        indice_0[i, 1] * Min_2[:, 1] +
                        indice_0[i, 2] * Min_2[:, 2])

    V1 = (V1[~np.all(V1 == 0, axis=1)]).astype(int)
    V2 = (V2[~np.all(V2 == 0, axis=1)]).astype(int)
    MeanPlanes = (V1 + V2) / 2

    # Check the type of GB plane: Symmetric tilt, tilt, twist
    for i in range(len(V1)):
        if ang(V1[i], uvw) < tol:

            for j in range(len(symmetry_equivalent(MP))):
                if 1 - ang(MeanPlanes[i], symmetry_equivalent(MP)[j]) < tol:
                    GBtype.append('Symmetric Tilt')
                    break
            else:
                GBtype.append('Tilt')
        elif 1 - ang(V1[i], uvw) < tol:
            GBtype.append('Twist')
        else:
            GBtype.append('Mixed')

    return (V1, V2, MeanPlanes, GBtype)


def find_orthogonal_cell(uvw, sigma, theta, R, m, n, GB1, 
    Min_1, Min_2, tol = 0.001):
    """
    finds Orthogonal cells from the CSL minimal cells.
    arguments:
    basis -- lattice basis
    uvw -- rotation axis
    m,n -- two necessary integers
    GB1 -- input plane orientation
    """
    # Inputs
    lim = 15
    uvw = np.array(uvw)
    Theta = theta
    Sigma = sigma
    
    #create GB2 from GB1
    GB2 = np.round((dot(R, GB1)), 6)
    x = np.arange(-lim, lim + 1, 1)
    y = x
    z = x
    
    indice = (np.stack(np.meshgrid(x, y, z)).T).reshape(len(x) ** 3, 3)
    indice_0 = indice[np.where(np.sum(abs(indice), axis=1) != 0)]
    indice_0 = indice_0[np.argsort(norm(indice_0, axis=1))]
    OrthoCell_1 = np.zeros([3, 3])
    OrthoCell_1[:, 0] = np.array(GB1)
    OrthoCell_2 = np.zeros([3, 3])
    OrthoCell_2[:, 0] = np.array(GB2)
    
    # Find Ortho vectors:
    if ang(OrthoCell_1[:, 0], uvw) < tol:
        OrthoCell_1[:, 1] = uvw
        OrthoCell_2[:, 1] = uvw
    else:
        for i in range(len(indice_0)):
            v1 = (indice_0[i, 0] * Min_1[:, 0] +
                  indice_0[i, 1] * Min_1[:, 1] +
                  indice_0[i, 2] * Min_1[:, 2])
            v2 = (indice_0[i, 0] * Min_2[:, 0] +
                  indice_0[i, 1] * Min_2[:, 1] +
                  indice_0[i, 2] * Min_2[:, 2])
            if ang(v1, OrthoCell_1[:, 0]) < tol:
                OrthoCell_1[:, 1] = v1
                OrthoCell_2[:, 1] = v2
                break
    OrthoCell_1[:, 2] = np.cross(OrthoCell_1[:, 0], OrthoCell_1[:, 1])
    OrthoCell_2[:, 2] = np.cross(OrthoCell_2[:, 0], OrthoCell_2[:, 1])

    if (common_divisor(OrthoCell_1[:, 2])[1] ==
            common_divisor(OrthoCell_2[:, 2])[1]):
        OrthoCell_1[:, 2] = common_divisor(OrthoCell_1[:, 2])[0]
        OrthoCell_2[:, 2] = common_divisor(OrthoCell_2[:, 2])[0]
    
    Volume_1 = (round(det(OrthoCell_1), 5))
    Volume_2 = (round(det(OrthoCell_2), 5))
    
    if Volume_1 == Volume_2:
        return True, OrthoCell_1.astype(float), OrthoCell_2.astype(float)

    return False, None, None


def get_tilt_twist_comp(v1, uvw, m, n, tol=0.001):
    """
    returns the tilt and twist components of a given GB plane.
    arguments:
    v1 -- given gb plane
    uvw -- axis of rotation
    m,n -- the two necessary integers
    """
    theta = get_cubic_theta(uvw, m, n)
    R = rot(uvw, theta)
    v2 = np.round(dot(R, v1), 6).astype(int)
    tilt = angv(v1, v2)
    twist = 0
    gb_type = ""
    if abs(tilt - degrees(theta)) < 10e-5:
        gb_type = "tilt"
    else:
        twist = 2 * acos(cos(theta / 2) / cos(radians(tilt / 2)))
    
    #Assign types
    MP = np.array([[1, 0, 0],
                   [0, 1, 0],
                   [0, 0, 1],
                   [1, 1, 0],
                   [0, 1, 1],
                   [1, 0, 1],
                   ], dtype='float')
        

    return tilt, twist

#Helpers

def common_divisor(a):
    """
    returns the common factor of vector a and the reduced vector.
    """
    CommFac = []
    a = np.array(a)
    for i in range(2, 100):
        while (a[0] % i == 0 and a[1] % i == 0 and a[2] % i == 0):
            a = a / i
            CommFac.append(i)
    return(a.astype(int), (abs(np.prod(CommFac))))

def integer_array(A, tol=1e-7):
    """
    returns True if an array is ineteger.
    """
    return np.all(abs(np.round(A) - A) < tol)

def integer_matrix(a):
    """
    returns an integer matrix from row vectors.
    """
    Found = True
    b = np.zeros((3, 3))
    a = np.array(a)
    for i in range(3):
        for j in range(1, 2000):
            testV = j * a[i]
            if integer_array(testV):
                b[i] = testV
                break
        if all(b[i] == 0):
            Found = False
            print("Can not make integer matrix!")
    return (b) if Found else None

def angv(a, b):
    """
    returns the angle between two vectors.
    """
    ang = acos(np.round(dot(a, b)/norm(a)/norm(b), 8))
    return round(degrees(ang), 7)


def ang(a, b):
    """
    returns the cos(angle) between two vectors.
    """
    ang = np.round(dot(a, b)/norm(a)/norm(b), 7)
    return abs(ang)

def symmetry_equivalent(arr):
    """
    returns cubic symmetric eqivalents of the given 2 dimensional vector.
    """
    Sym = np.zeros([24, 3, 3])
    Sym[0, :] = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    Sym[1, :] = [[1, 0, 0], [0, -1, 0], [0, 0, -1]]
    Sym[2, :] = [[-1, 0, 0], [0, 1, 0], [0, 0, -1]]
    Sym[3, :] = [[-1, 0, 0], [0, -1, 0], [0, 0, 1]]
    Sym[4, :] = [[0, -1, 0], [-1, 0, 0], [0, 0, -1]]
    Sym[5, :] = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]
    Sym[6, :] = [[0, 1, 0], [-1, 0, 0], [0, 0, 1]]
    Sym[7, :] = [[0, 1, 0], [1, 0, 0], [0, 0, -1]]
    Sym[8, :] = [[-1, 0, 0], [0, 0, -1], [0, -1, 0]]
    Sym[9, :] = [[-1, 0, 0], [0, 0, 1], [0, 1, 0]]
    Sym[10, :] = [[1, 0, 0], [0, 0, -1], [0, 1, 0]]
    Sym[11, :] = [[1, 0, 0], [0, 0, 1], [0, -1, 0]]
    Sym[12, :] = [[0, 1, 0], [0, 0, 1], [1, 0, 0]]
    Sym[13, :] = [[0, 1, 0], [0, 0, -1], [-1, 0, 0]]
    Sym[14, :] = [[0, -1, 0], [0, 0, 1], [-1, 0, 0]]
    Sym[15, :] = [[0, -1, 0], [0, 0, -1], [1, 0, 0]]
    Sym[16, :] = [[0, 0, 1], [1, 0, 0], [0, 1, 0]]
    Sym[17, :] = [[0, 0, 1], [-1, 0, 0], [0, -1, 0]]
    Sym[18, :] = [[0, 0, -1], [1, 0, 0], [0, -1, 0]]
    Sym[19, :] = [[0, 0, -1], [-1, 0, 0], [0, 1, 0]]
    Sym[20, :] = [[0, 0, -1], [0, -1, 0], [-1, 0, 0]]
    Sym[21, :] = [[0, 0, -1], [0, 1, 0], [1, 0, 0]]
    Sym[22, :] = [[0, 0, 1], [0, -1, 0], [1, 0, 0]]
    Sym[23, :] = [[0, 0, 1], [0, 1, 0], [-1, 0, 0]]

    arr = np.atleast_2d(arr)
    Result = []
    for i in range(len(Sym)):
        for j in range(len(arr)):
            Result.append(dot(Sym[i, :], arr[j]))
    Result = np.array(Result)
    return np.unique(Result, axis=0)


# GB building tools
def generate_ortho_unitcell_atoms(ortho, basis):

    """
    populates a unitcell from the orthogonal vectors.
    """
    Or = ortho.T
    Orint = integer_matrix(Or)
    LoopBound = np.zeros((3, 2), dtype=float)
    transformed = []
    CubeCoords = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 0],
                          [0, 1, 1], [1, 0, 1], [1, 1, 1], [0, 0, 0]],
                          dtype=float)
    for i in range(len(CubeCoords)):
        transformed.append(np.dot(Orint.T, CubeCoords[i]))

    # Finding bounds for atoms in a CSL unitcell:
    LoopBound[0, :] = [min(np.array(transformed)[:, 0]),
                       max(np.array(transformed)[:, 0])]
    LoopBound[1, :] = [min(np.array(transformed)[:, 1]),
                       max(np.array(transformed)[:, 1])]
    LoopBound[2, :] = [min(np.array(transformed)[:, 2]),
                       max(np.array(transformed)[:, 2])]

    # Filling up the unitcell:
    Tol = 1
    x = np.arange(LoopBound[0, 0] - Tol, LoopBound[0, 1] + Tol + 1, 1)
    y = np.arange(LoopBound[1, 0] - Tol, LoopBound[1, 1] + Tol + 1, 1)
    z = np.arange(LoopBound[2, 0] - Tol, LoopBound[2, 1] + Tol + 1, 1)
    V = len(x) * len(y) * len(z)
    indice = (np.stack(np.meshgrid(x, y, z)).T).reshape(V, 3)
    Base = basis
    Atoms = []
    tol = 0.001
    if V > 5e6:
        print("Warning! It may take a very long time"
              "to produce this cell!")
    # produce Atoms:

    for i in range(V):
        for j in range(len(Base)):
            Atoms.append(indice[i, 0:3] + Base[j, 0:3])
    Atoms = np.array(Atoms)

    # Cell conditions
    Con1 = dot(Atoms, Or[0]) / norm(Or[0]) + tol
    Con2 = dot(Atoms, Or[1]) / norm(Or[1]) + tol
    Con3 = dot(Atoms, Or[2]) / norm(Or[2]) + tol
    # Application of the conditions:
    Atoms = (Atoms[(Con1 >= 0) & (Con1 <= norm(Or[0])) & (Con2 >= 0) &
             (Con2 <= norm(Or[1])) &
             (Con3 >= 0) & (Con3 <= norm(Or[2]))])

    if len(Atoms) == (round(det(Or) * len(Base), 7)).astype(int):
        return Atoms

def generate_bicrystal_atoms(ortho1, ortho2, basis):
    """
    builds the unitcells for both grains g1 and g2.
    """
    Or_1 = ortho1.T
    Or_2 = ortho2.T
    rot1 = np.array([Or_1[0, :] / norm(Or_1[0, :]),
                         Or_1[1, :] / norm(Or_1[1, :]),
                         Or_1[2, :] / norm(Or_1[2, :])])
    rot2 = np.array([Or_2[0, :] / norm(Or_2[0, :]),
                         Or_2[1, :] / norm(Or_2[1, :]),
                         Or_2[2, :] / norm(Or_2[2, :])])

    atoms1 = generate_ortho_unitcell_atoms(ortho1.copy(), basis)
    atoms2 = generate_ortho_unitcell_atoms(ortho2.copy(), basis)

    atoms1 = dot(rot1, atoms1.T).T
    atoms2 = dot(rot2, atoms2.T).T
    
    atoms2[:, 0] = atoms2[:, 0] - norm(Or_2[0, :])  # - tol
    return atoms1, atoms2


def expand_super_cell(ortho1, atoms1, atoms2, dim):
    """
    expands the smallest CSL unitcell to the given dimensions.
    """
    a = norm(ortho1[:, 0])
    b = norm(ortho1[:, 1])
    c = norm(ortho1[:, 2])
    dimX, dimY, dimZ = dim

    X = atoms1.copy()
    Y = atoms2.copy()

    X_new = []
    Y_new = []
    for i in range(dimX):
        for j in range(dimY):
            for k in range(dimZ):
                Position1 = [i * a, j * b, k * c]
                Position2 = [-i * a, j * b, k * c]
                for l in range(len(X)):
                    X_new.append(Position1[0:3] + X[l, 0:3])
                for m in range(len(Y)):
                    Y_new.append(Position2[0:3] + Y[m, 0:3])

    atoms1 = np.array(X_new)
    atoms2 = np.array(Y_new)
    return atoms1, atoms2

def find_overlapping_atoms(ortho1, atoms1, atoms2, dim, overlap=0.0):
    """
    finds the overlapping atoms.
    """
    periodic_length = norm(ortho1[:, 0]) * dim[0]
    periodic_image = atoms2 + [periodic_length * 2, 0, 0]
    # select atoms contained in a smaller window around the GB and its
    # periodic image
    IndX = np.where([(atoms1[:, 0] < 1) |
                     (atoms1[:, 0] > (periodic_length - 1))])[1]
    IndY = np.where([atoms2[:, 0] > -1])[1]
    IndY_image = np.where([periodic_image[:, 0] <
                          (periodic_length + 1)])[1]
    X_new = atoms1[IndX]
    Y_new = np.concatenate((atoms2[IndY], periodic_image[IndY_image]))
    IndY_new = np.concatenate((IndY, IndY_image))
    # create a meshgrid search routine
    x = np.arange(0, len(X_new), 1)
    y = np.arange(0, len(Y_new), 1)
    indice = (np.stack(np.meshgrid(x, y)).T).reshape(len(x) * len(y), 2)
    norms = norm(X_new[indice[:, 0]] - Y_new[indice[:, 1]], axis=1)
    indice_x = indice[norms < overlap][:, 0]
    indice_y = indice[norms < overlap][:, 1]
    X_del = X_new[indice_x]
    Y_del = Y_new[indice_y]

    if (len(X_del) != len(Y_del)):
        print("Warning! the number of deleted atoms"
              "in the two grains are not equal!")
    # print(type(IndX), len(IndY), len(IndY_image))
    return (X_del, Y_del, IndX[indice_x], IndY_new[indice_y])

def populate_gb(ortho1, ortho2, basis, 
    lattice_parameter, dim=(1,1,1), 
    overlap=0.0, rigid=False):
    
    atoms1, atoms2 = generate_bicrystal_atoms(ortho1, ortho2, basis)
    if overlap > 0.0:
        which_gb = "g1"
    elif overlap < 0.0:
        which_gb = "g2"
    else:
        which_gb = "a"

    atoms1, atoms2 = expand_super_cell(ortho1, atoms1, atoms2, dim)
    xdel, _, x_indice, y_indice = find_overlapping_atoms(ortho1, atoms1, atoms2, dim, overlap=overlap)

    if which_gb == "g1":
        atoms1 = np.delete(atoms1, x_indice, axis=0)
    elif which_gb == "g2":
        atoms2 = np.delete(atoms2, y_indice, axis=0)

    atoms1 = atoms1*lattice_parameter
    atoms2 = atoms2*lattice_parameter
    dimx, dimy, dimz = dim

    #get box bounds
    xlo = -1 * np.round(norm(ortho1[:, 0]) * dimx * lattice_parameter, 8)
    xhi = np.round(norm(ortho1[:, 0]) * dimx * lattice_parameter, 8)
    LenX = xhi - xlo
    ylo = 0.0
    yhi = np.round(norm(ortho1[:, 1]) * dimy * lattice_parameter, 8)
    LenY = yhi - ylo
    zlo = 0.0
    zhi = np.round(norm(ortho1[:, 2]) * dimz * lattice_parameter, 8)
    LenZ = zhi - zlo    

    return [[LenX, 0, 0],[0, LenY, 0],[0, 0, LenZ]], atoms1, atoms2





import numpy as np
import gzip
from ase import Atom, Atoms
import gzip
import io
import os
from ase.io import write, read
import pyscal.formats.ase as ptase
import warnings

def read_snap(infile, compressed = False):
    """
    Function to read a POSCAR format.

    Parameters
    ----------
    infile : string
        name of the input file

    compressed : bool, optional
        force to read a `gz` zipped file. If the filename ends with `.gz`, use of this keyword is not
        necessary, Default False

    Returns
    -------
    atoms : list of `Atom` objects
        list of all atoms as created by user input

    box : list of list of floats
        list of the type `[[xlow, xhigh], [ylow, yhigh], [zlow, zhigh]]` where each of them are the lower
        and upper limits of the simulation box in x, y and z directions respectively.

    Examples
    --------
    >>> atoms, box = read_poscar('POSCAR')
    >>> atoms, box = read_poscar('POSCAR.gz')
    >>> atoms, box = read_poscar('POSCAR.dat', compressed=True)

    """
    aseobj = read(infile, format="vasp")
    atoms, box = ptase.read_snap(aseobj)
    return atoms, box


def write_snap(sys, outfile, comments="pyscal", species=None):
    """
    Function to read a POSCAR format.

    Parameters
    ----------
    outfile : string
        name of the input file


    """
    if species is None:
        warnings.warn("Using legacy poscar writer, to use ASE backend specify species")
        write_poscar(sys, outfile, comments=comments)
    else:
        aseobj = ptase.convert_snap(sys, species=species)
        write(outfile, aseobj, format="vasp")


def split_snaps(**kwargs):
    raise NotImplementedError("split method for mdtraj is not implemented")

def convert_snap(**kwargs):
    raise NotImplementedError("convert method for mdtraj is not implemented")

def write_poscar(sys, outfile, comments="pyscal"):
    """
    Function to read a POSCAR format.
    Parameters
    ----------
    outfile : string
        name of the input file
    """

    fout = open(outfile, 'w')

    fout.write(comments+"\n")
    fout.write("   1.00000000000000\n")

    #write box
    vecs = sys.box
    fout.write("      %1.14f %1.14f %1.14f\n"%(vecs[0][0], vecs[0][1], vecs[0][2]))
    fout.write("      %1.14f %1.14f %1.14f\n"%(vecs[1][0], vecs[1][1], vecs[1][2]))
    fout.write("      %1.14f %1.14f %1.14f\n"%(vecs[2][0], vecs[2][1], vecs[2][2]))

    atypes = sys.atoms.types
    
    tt, cc  = np.unique(atypes, return_counts=True)
    
    atomgroups = [[] for x in range(len(tt))]
    
    for count, t in enumerate(tt):
        for cc, pos in enumerate(sys.atoms.positions):
            if int(atypes[cc]) == t:
                atomgroups[count].append(pos)

    fout.write("  ")
    for c in cc:
        fout.write("%d   "%int(c))
    fout.write("\n")

    fout.write("Cartesian\n")

    for i in range(len(atomgroups)):
        for pos in atomgroups[i]:
            fout.write(" %1.14f %1.14f %1.14f\n"%(pos[0], pos[1], pos[2]))

    fout.close()



import pytest
import numpy as np
import pyscal.core as pc
import pyscal.crystal_structures as pcs

def test_ordered_disorder():
    sys = pc.System('tests/files/conf.fcc.dump')
    sys.find_neighbors(method='cutoff', cutoff=0)
    sys.calculate_q(6)
    sys.calculate_disorder(averaged=True)
    assert np.mean(sys.atoms.steinhardt.disorder.norm) < 0.50
    assert np.mean(sys.atoms.steinhardt.disorder.average) < 0.50
    
def test_disordered_disorder():
    sys = pc.System('tests/files/conf.lqd.dump')
    sys.find_neighbors(method='cutoff', cutoff=0)
    sys.calculate_q(6)
    sys.calculate_disorder(averaged=True)
    assert np.mean(sys.atoms.steinhardt.disorder.norm) > 1.00
    assert np.mean(sys.atoms.steinhardt.disorder.average) > 1.00



import numpy as np
import os
import pyscal.csl as csl
import pyscal.crystal_structures as pcs
from pyscal.core import System, Atoms

class GrainBoundary:
    def __init__(self):
        self._axis = None
        self.limit = None
        self._sigma = None
        self._theta = None
        self.gb_index_limit = None
        self.gb_plane = None
        
        
    @property
    def axis(self):
        return self._axis
    
    @axis.setter
    def axis(self, axis):
        self._axis = np.array(axis)
    
    @property
    def sigma(self):
        return self._sigma
    
    @sigma.setter
    def sigma(self, val):
        self._sigma = val
    
    def get_possible_sigma(self):
        if self._axis is not None:
            sigmas, thetas = csl.get_sigma_list(self._axis, self.limit)
            return sigmas
        
    #theta is calculated; not set
    @property
    def theta(self):
        return np.degrees(self._theta)
    
    def calculate_theta(self):
        theta, m, n = csl.get_theta_m_n_list(self._axis, self._sigma)[0]
        R = csl.rot(self._axis, theta)
        M1, M2 = csl.create_minimal_cell_method_1(self._sigma, self._axis, R)
        self._theta = theta
        self._m = m
        self._n = n
        self._M1 = M1
        self._M2 = M2
        self._R = R
        return self._theta
    
    def calculate_possible_gb_planes(self):
        V1, V2, M, Gb = csl.create_possible_gb_plane_list(self._axis, 
                            self._m, self._n, self.gb_index_limit)
        return V1, V2, M, Gb

    def check_if_gb_is_valid(self, gb_vector):
        if self._theta is None:
            self.calculate_theta()
        validity, ortho_1, ortho_2 = csl.find_orthogonal_cell(self._axis, self._sigma, self._theta, 
                                                                   self._R, self._m, self._n, gb_vector,
                                                                   self._M1, self._M2, tol = 0.001)
        self._ortho_1 = ortho_1
        self._ortho_2 = ortho_2
        return validity
    
    def find_gb_character(self):
        if self.gb_plane is not None:
            self.gb_index_limit = max(np.abs(self.gb_plane)+1)
            outs = self.calculate_possible_gb_planes()
            main_str_array = np.array([" ".join(np.array(x).astype(str)) for x in outs[0]])
            search_array = " ".join(np.array(self.gb_plane).astype(str))
            arg = np.argwhere(main_str_array == search_array)
            return outs[3][arg[0][0]]

    def create_grain_boundary(self, axis, 
                              sigma, 
                              gb_plane,
                              limit=3, 
                              gb_index_limit=10,
                              ):
        self.axis = axis
        self.sigma = sigma
        self.limit = limit
        self.gb_plane = gb_plane
        self.gb_index_limit = gb_index_limit
        valid = self.check_if_gb_is_valid(gb_plane)
        if not valid:
            raise TypeError("GB cannot be created with the given input!")
        return valid
    
    def populate_grain_boundary(self, lattice, element=None, repetitions=(1,1,1), lattice_parameter=1, overlap=0.0):
        if lattice in pcs.structures.keys():
            structure = lattice
            element = element
            basis = pcs.structures[lattice]['positions']
            sdict = pcs.structures[lattice]
        elif lattice in pcs.elements.keys():
            structure = pcs.elements[lattice]['structure']
            element = lattice
            lattice_parameter = pcs.elements[lattice]['lattice_constant']
            basis = pcs.structures[structure]['positions']
            sdict = pcs.structures[structure]
        else:
            raise ValueError("Unknown lattice type")
        box, atoms1, atoms2 = csl.populate_gb(self._ortho_1, self._ortho_2, 
                        np.array(basis),
                        lattice_parameter,
                        dim=repetitions, 
                        overlap=overlap)
        atoms = Atoms()
        total_atoms = np.concatenate((atoms1, atoms2))
        if element is not None:
            atoms.from_dict({"positions": total_atoms, "species": [element for x in range(len(total_atoms))]})
        else:
            atoms.from_dict({"positions": total_atoms})
        sys = System()
        sys.box = box
        sys.atoms = atoms
        sys.atoms._lattice = structure
        sys.atoms._lattice_constant = lattice_parameter
        sys._structure_dict = sdict
        sys.remap_atoms_into_box()
        return sys


"""
pyscal module for creating crystal structures.
"""

import numpy as np
import warnings
import os
from functools import partial
from functools import update_wrapper


from pyscal.attributes import read_yaml
from pyscal.core import System
from pyscal.structure_creator import make_crystal

structures = read_yaml(os.path.join(os.path.dirname(__file__), "data/structure_data.yaml"))
elements = read_yaml(os.path.join(os.path.dirname(__file__), "data/element_data.yaml"))


#wrapper methods
def structure_creator(structure, 
             lattice_constant = 1.00, 
             repetitions = None, 
             ca_ratio = 1.633, 
             noise = 0,
             element=None):
    """
    Create a crystal structure and return it as a System object.

    Parameters
    ----------
    structure : {'sc', 'bcc', 'fcc', 'hcp', 'diamond', 'a15' or 'l12'}
        type of the crystal structure

    lattice_constant : float, optional
        lattice constant of the crystal structure, default 1

    repetitions : list of ints of len 3, optional
        of type `[nx, ny, nz]`, repetions of the unit cell in x, y and z directions.
        default `[1, 1, 1]`.

    ca_ratio : float, optional
        ratio of c/a for hcp structures, default 1.633

    noise : float, optional
        If provided add normally distributed noise with standard deviation `noise` to the atomic positions.
    
    element : string, optional
        The chemical element

    Returns
    -------
    System: pyscal System
        system will be populated with given atoms and simulation box

    Examples
    --------
    >>> sys = structure_creator('bcc', lattice_constant=3.48, repetitions=[2,2,2])
    """    
    return System.from_structure(structure, lattice_constant=lattice_constant,
             repetitions=repetitions, ca_ratio=ca_ratio,
             noise=noise, element=element)

class Structure:
    """
    A class for structure creation
    
    Attributes
    ----------
    element: 
        create elementary structures
    lattice:
        create structures by specifying lattice
    """
    def __init__(self):
        #create by element name
        self.element = ElementCreator(elements)
        #create by lattice name
        self.lattice = LatticeCreator(structures)
        #create a general structure
        self.custom = general_lattice
        #complete dict
        self._structure_dict = structures

    def structure_dict(self, structure):
        return self._structure_dict[structure]

class ElementCreator:
    """
    Create an elementary structure
    """
    def __init__(self, element_dict):
        self._element_dict = element_dict
    
    def __dir__(self):
        return list(self._element_dict.keys())
    
    def __getattr__(self, key):
        #this is the element based creater
        if key in self._element_dict.keys():
            structure = self._element_dict[key]['structure']
            pfunc = partial(structure_creator, structure,
                        lattice_constant=self._element_dict[key]['lattice_constant'],
                        element = key)
            update_wrapper(pfunc, structure_creator)
            return pfunc

class LatticeCreator(ElementCreator):
    """
    Create a lattice
    """
    def __getattr__(self, key):
        if key in self._element_dict.keys():
            pfunc = partial(structure_creator, key)
            update_wrapper(pfunc, structure_creator)
            return pfunc


#general structure creator
def general_lattice(species, positions, 
    scaling_factors=[1.0, 1.0, 1.0],
    lattice_constant = 1.00, 
    repetitions = None, 
    noise = 0,
    element=None):
    """
    Create a general lattice structure.

    species: list
        list of per-atom species

    positions:
        list of relative positions positions of reach atom (between 0-1)

    scaling_fractors:
        factors with which the unit cell should be scaled, for example hcp could
        have [1,1.73, 1.63]. Default [1,1,1]

    lattice_constant : float, optional
        lattice constant of the crystal structure, default 1

    repetitions : list of ints of len 3, optional
        of type `[nx, ny, nz]`, repetions of the unit cell in x, y and z directions.
        default `[1, 1, 1]`.

    noise : float, optional
        If provided add normally distributed noise with standard deviation `noise` to the atomic positions.
    
    element : string, optional
        The chemical element
    """
    if not (len(species) == len(positions)):
        raise ValueError("Species and positions should have same length!")

    sdict = {"custom":
                {"natoms": len(positions),
                 "species": species,
                 "scaling_factors": scaling_factors,
                 "positions": positions}
            }

    atoms, box = make_crystal("custom", lattice_constant=lattice_constant,
        repetitions=repetitions, noise=noise, element=element,
        structures=sdict)

    sys = System()
    sys.box = box
    sys.atoms = atoms
    sys.atoms._lattice = None
    sys.atoms._lattice_constant = lattice_constant
    sys._structure_dict = sdict["custom"]
    return sys

        

    
    




"""


"""


import os
import numpy as np
import warnings
import itertools
from ase.io import write
import uuid
import gzip
import io
from scipy.special import sph_harm
import copy

from pyscal.atoms import Atoms, AttrSetter
import pyscal.csystem as pc
import pyscal.traj_process as ptp
from pyscal.formats.ase import convert_snap
import pyscal.structure_creator as pcs
import pyscal.operations.operations as po
#import pyscal.routines as routines
#import pyscal.visualization as pv


class System:
    """
    Python class for holding the properties of an atomic configuration 
    """    
    def __init__(self, filename=None, format="lammps-dump", 
                                            compressed = False, customkeys=None):
        self.initialized = True
        self.neighbors_found = False
        self.neighbor_method = None
        self.ghosts_created = False
        self.actual_box = None

        #box parameters
        self.rot = [[0,0,0], [0,0,0], [0,0,0]]
        self.rotinv = [[0,0,0], [0,0,0], [0,0,0]]
        self.boxdims = [0,0,0]
        self.triclinic = 0
        self._structure_dict = None
        self._atoms = Atoms()
        
        if filename is not None:
            self.read_inputfile(filename, format=format, 
                                            compressed = compressed, customkeys=customkeys)
    
    @classmethod
    def from_structure(cls, structure, lattice_constant = 1.00, repetitions = None, ca_ratio = 1.633, noise = 0, element=None, chemical_symbol=None):
        atoms, box, sdict = pcs.make_crystal(structure, lattice_constant=lattice_constant,
             repetitions=repetitions, ca_ratio=ca_ratio,
             noise=noise, element=element, return_structure_dict=True)
        obj = cls()
        obj.box = box
        obj.atoms = atoms
        obj.atoms._lattice = structure
        obj.atoms._lattice_constant = lattice_constant
        obj._structure_dict = sdict
        return obj

    def iter_atoms(self):
        return self.atoms.iter_atoms()

    @property
    def natoms(self):
        return self.atoms.natoms

    @property
    def concentration(self):
        return self.atoms.composition

    @property
    def composition(self):
        return self.atoms.composition

    @property
    def box(self):
        """
        Wrap for inbuilt box
        """
        if self.actual_box is not None:
            return self.actual_box
        else:
            return self._box

    @box.setter
    def box(self, userbox):
        """
        Box setter 
        """
        #we should automatically check for triclinic cells here
        summ = 0
        for i in range(3):
            box1 = np.array(userbox[i-1])
            box2 = np.array(userbox[i])
            summ += np.dot(box1, box2)/(np.linalg.norm(box1)*np.linalg.norm(box2))

        #check if the summ is zero
        if (np.abs(summ) > 1E-6):
            #this is a triclinic box
            rot = np.array(userbox).T
            rotinv = np.linalg.inv(rot)
            self.triclinic = 1
            self.rot = rot
            self.rotinv = rotinv

        self.boxdims[0] = np.sum(np.array(userbox[0])**2)**0.5
        self.boxdims[1] = np.sum(np.array(userbox[1])**2)**0.5
        self.boxdims[2] = np.sum(np.array(userbox[2])**2)**0.5
        self._box = userbox

    @property
    def box_dimensions(self):
        return [np.linalg.norm(self.box[x]) for x in range(3)]

    @property
    def direct_coordinates(self):
        dims = self.box_dimensions
        xfrac = np.array(self.atoms.positions)[:,0]/dims[0]
        yfrac = np.array(self.atoms.positions)[:,1]/dims[1]
        zfrac = np.array(self.atoms.positions)[:,2]/dims[2]
        coords = np.column_stack((xfrac, yfrac, zfrac))
        return coords    
    
    @property
    def volume(self):
        """
        Volume of box
        """
        vol = np.dot(np.cross(self._box[0], self._box[1]), self._box[2])
        return vol

    @property
    def atoms(self):
        return self._atoms
    
    @atoms.setter
    def atoms(self, atoms):
        """
        Set atoms
        """
        if(len(atoms['positions']) < 200):
            #we need to estimate a rough idea
            needed_atoms = 200 - len(atoms)
            #get a rough cell
            needed_cells = np.ceil(needed_atoms/len(atoms))
            nx = int(needed_cells**(1/3))
            nx = int(np.ceil(nx/2))

            if np.sum(self.box) == 0:
                raise ValueError("Simulation box should be initialized before atoms")
            atoms, box = self.repeat((nx, nx, nx), atoms=atoms, ghost=True, scale_box=True, assign=False, return_atoms=True)
            self.actual_box = self.box.copy()
            self.box = box

        self._atoms = atoms
        


    def add_atoms(self, atoms):
        """
        Cleanly add a given list of atoms

        Parameters
        ----------
        atoms : dict

        Returns
        -------
        None
        """ 
        ## MOVE TO ATOMS
        self._atoms.add_atoms(atoms)

    def repeat(self, repetitions, atoms=None, ghost=False, scale_box=True, assign=False, return_atoms=False):
        """
        """
        return po.repeat(self, repetitions, 
            atoms=atoms, ghost=ghost, 
            scale_box=scale_box, 
            return_atoms=return_atoms)

    def remap_atoms_into_box(self):
        """
        Go through atoms in the list and remap them into the bix
        """
        rot = np.array(self.box).T
        rotinv = np.linalg.inv(rot)

        for x in range(self.natoms):
            pos = pc.remap_atom_into_box(self.atoms["positions"][x], 
                self.triclinic,
                rot, 
                rotinv, 
                self.box_dimensions)
            self.atoms["positions"][x] = pos        

    def apply_mask(self, mask_type="primary", ids=None, indices=None, condition=None, selection=False):
        """

        Notes
        -----
        Masks can be used to exclude atoms from neighbor calculations. An atom for which
        mask is set to True is excluded from the calculation. There are two types of masks,
        `primary` or `secondary`. For example, neighbors are being calculated for a central
        atom `i`. The neighbor atom is denoted as `j`. If `primary` mask of `i` is True, no neighbor
        calculation is carried out for `i`. If it is False, `i` is considered. Now if `secondary`
        mask of `j` is True, it will not included in the list of neighbors of `i` even if it is within
        the cutoff distance. The `primary` mask of `j` has no effect in this situation.

        An example situation can be to calculate the local concentration around Ni atoms in a NiAl
        structure. In this case, the `primary` mask of all Al atoms can be set to True so that
        only `Ni` atoms are considered. Now, in a second case, the task is to count the number of Al
        atoms around each Ni atom. For this case, the `primary` mask of all Al atoms can be set to True,
        and the `secondary` mask of all Ni atoms can be set to True.

        The masks for ghost atoms are copied from the corresponding mask for real atoms.
        """
        #check if length of mask is equal to length of real atoms
        self.atoms.apply_mask(mask_type=mask_type, ids=ids, 
            indices=indices, condition=condition, selection=selection)

    def remove_mask(self, mask_type="primary", ids=None, indices=None, condition=None, selection=False):
        """
        Remove applied masks

        Parameters
        ----------
        mask_type: string, optional
            type of mask to be applied, either `primary`, `secondary` or `all`

        Returns
        -------
        None
        """
        self._atoms.remove_mask(mask_type=mask_type, ids=ids, 
            indices=indices, condition=condition, selection=selection)

    def apply_selection(self, ids=None, indices=None, condition=None):
        self._atoms.apply_selection(ids=ids, indices=indices, condition=condition)    
    
    def remove_selection(self, ids=None, indices=None, condition=None):
        self._atoms.remove_selection(ids=ids, indices=indices, condition=condition)
    
    def delete(self, ids=None, indices=None, condition=None, selection=False):
        self._atoms.delete(ids=ids, indices=indices, condition=condition, selection=selection)


    def embed_in_cubic_box(self, input_box=None, return_box=False):
        """
        Embedded the triclinic box in a cubic box
        """
        return po.embed_in_cubic_box(self, input_box=input_box,
            return_box=return_box) 

    def get_distance(self, pos1, pos2, vector=False):
        """
        Get the distance between two atoms.

        Parameters
        ----------
        pos1 : list
                first atom position
        pos2 : list
                second atom position
        vector: bool, optional
            If True, return the vector between two atoms

        Returns
        -------
        distance : double
                distance between the first and second atom.

        Notes
        -----
        Periodic boundary conditions are assumed by default.
        """
        
        if vector:
            return dist, diff
        else:
            return dist

    def get_concentration(self):
        """
        Return a dict containing the concentration of the system

        Parameters
        ----------
        None

        Returns
        -------
        condict : dict
            dict of concentration values
        """
        return self.concentration


    def read_inputfile(self, filename, format="lammps-dump", 
                                            compressed = False, customkeys=None):
        """

        Read input file that contains the information of system configuration.

        Parameters
        ----------
        filename : string
            name of the input file.

        format : {'lammps-dump', 'poscar', 'ase', 'mdtraj'}
            format of the input file, in case of `ase` the ASE Atoms object

        compressed : bool, optional
            If True, force to read a `gz` compressed format, default False.

        customkeys : list
            A list containing names of headers of extra data that needs to be read in from the
            input file.

        Returns
        -------
        None

        Notes
        -----
        `format` keyword specifies the format of the input file. Currently only
        a `lammps-dump` and `poscar` files are supported.  Additionaly, the widely
        use Atomic Simulation environment (https://wiki.fysik.dtu.dk/ase/ase/ase.html).
        mdtraj objects (http://mdtraj.org/1.9.3/) are also supported by using the keyword
        `'mdtraj'` for format. Please note that triclinic boxes are not yet supported for
        mdtraj format.
        Atoms object can also be used directly. This function uses the
        :func:`~pyscal.traj_process` module to process a file which is then assigned to system.

        `compressed` keyword is not required if a file ends with `.gz` extension, it is
        automatically treated as a compressed file.

        Triclinic simulation boxes can also be read in.

        If `custom_keys` are provided, this extra information is read in from input files if
        available. This information is can be accessed directly as `self.atoms['customkey']`


        """
        self.neighbors_found = False
        self.neighbor_method = None
        self.ghosts_created = False
        self.actual_box = None

        #box parameters
        self.rot = [[0,0,0], [0,0,0], [0,0,0]]
        self.rotinv = [[0,0,0], [0,0,0], [0,0,0]]
        self.boxdims = [0,0,0]
        self.triclinic = 0

        atoms, box = ptp.read_file(filename, format=format, 
                                    compressed=compressed, customkeys=customkeys,)
        self.box = box
        self.atoms = atoms

    def to_file(self, outfile, format='lammps-dump', customkeys=None, customvals=None,
                compressed=False, timestep=0, species=None):
        """
        Save the system instance to a trajectory file.

        Parameters
        ----------
        outfile : string
            name of the output file

        format : string, {'lammps-dump', 'lammps-data', 'poscar'}
            format of the output file, default `lammps-dump`
            Currently only `lammps-dump` format is supported.

        customkeys : list of strings, optional
            a list of extra atom wise values to be written in the output file.

        customvals : list or list of lists, optional
            If `customkey` is specified, `customvals` take an array of the same length
            as number of atoms, which contains the values to be written out.

        compressed : bool, optional
            If true, the output is written as a compressed file.

        timestep : int, optional
            timestep to be written to file. default 0

        species : None, optional
            species of the atoms. Required if any format other than 'lammps-dump' is used. Required
            for convertion to ase object.

        Returns
        -------
        None

        Notes
        -----
        `to_file` method can handle a number of file formats. The most customizable format is the
        `lammps-dump` which can take a custom options using customkeys and customvals. customkeys
        will be the header written to the dump file. It can be any Atom attribute, any property
        stored in custom variable of the Atom, or calculated q values which can be given by `q4`,
        `aq4` etc. External values can also be provided using `customvals` option. `customvals` array
        should be of the same length as the number of atoms in the system.

        For all other formats, ASE is used to write out the file, and hence the `species` keyword
        needs to be specified. If initially, an ASE object was used to create the System, `species`
        keyword will already be saved, and need not be specified. In other cases, `species` should
        be a list of atomic species in the System. For example `["Cu"]` or `["Cu", "Al"]`, depending
        on the number of species in the System. In the above case, atoms of type 1 will be mapped to
        Cu and of type 2 will be mapped to Al. For a complete list of formats that ASE can handle,
        see `here <https://wiki.fysik.dtu.dk/ase/ase/io/io.html>`_ . 
        """

        ptp.write_file(self, outfile, format = format,
            compressed = compressed, customkeys = customkeys, customvals = customvals,
            timestep = timestep, species = species)

    def to_ase(self, species=None):
        """
        Convert system to an ASE Atoms object

        Parameters
        ----------
        species : list of string
            The chemical species

        Returns
        -------
        None
        """
        return convert_snap(self, species=species)

    def reset_neighbors(self):
        """
        Reset the neighbors of all atoms in the system.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        It is used automatically when neighbors are recalculated.

        """
        self.atoms["neighbors"] = []
        self.atoms["neighbordist"] = []
        self.atoms["temp_neighbors"] = []
        self.atoms["temp_neighbordist"] = []
        self.atoms["neighborweight"] = []
        self.atoms["diff"] = []
        self.atoms["r"] = []
        self.atoms["theta"] = []
        self.atoms["phi"] = []
        self.atoms["cutoff"] = []
        self.neighbors_found = False

        
        mapdict = {}
        mapdict["neighbors"] = {}
        mapdict["neighbors"]["index"] = "neighbors"
        mapdict["neighbors"]["distance"] = "neighbordist"
        mapdict["neighbors"]["weight"] = "neighborweight"
        mapdict["neighbors"]["displacement"] = "diff"
        mapdict["neighbors"]["cutoff"] = "cutoff"

        mapdict["neighbors"]["angle"] = {}
        mapdict["neighbors"]["angle"]["polar"] = "theta"
        mapdict["neighbors"]["angle"]["azimuthal"] = "phi"

        mapdict["neighbors"]["temporary"] = {}
        mapdict["neighbors"]["temporary"]["index"] = "temp_neighbors"
        mapdict["neighbors"]["temporary"]["distance"] = "temp_neighbordist"

        self.atoms._add_attribute(mapdict)

    def _check_neighbors(self):
        """
        Check if neighbors are calculated
        """
        if not self.neighbors_found:
            raise ValueError("This calculation needs neighbors to be calculated")

    def average_over_neighbors(self, key, include_self=True):
        """
        Perform a simple average over neighbor atoms

        Parameters
        ----------
        key: string
            atom property

        include_self: bool, optional
            If True, include the host atom in the calculation

        Returns
        -------

        """
        print(key)

        self._check_neighbors()

        if not key in self.atoms.keys():
            raise KeyError("required property not found!")

        test = self.atoms[key][0]

        if isinstance(test, list):
            raise TypeError("Averaging can only be done over 1D quantities")

        avgarr = []
        for i in range(len(self.atoms["positions"])):
            arr = []
            if include_self:
                arr.append(self.atoms[key][i])
            for j in self.atoms["neighbors"][i]:
                arr.append(self.atoms[key][j])
            avgarr.append(np.mean(arr))
        
        return avgarr 

    def find_neighbors(self, method='cutoff', cutoff=0, threshold=2, 
            voroexp=1, padding=1.2, nlimit=6, 
            cells=None, nmax=12, assign_neighbor=True):
        """

        Find neighbors of all atoms in the :class:`~pyscal.core.System`.

        Parameters
        ----------
        method : {'cutoff', 'voronoi', 'number'}
            `cutoff` method finds neighbors of an atom within a specified or adaptive cutoff distance from the atom.
            `voronoi` method finds atoms that share a Voronoi polyhedra face with the atom. Default, `cutoff`
            `number` method finds a specified number of closest neighbors to the given atom. Number only populates
            

        cutoff : { float, 'sann', 'adaptive'}
            the cutoff distance to be used for the `cutoff` based neighbor calculation method described above.
            If the value is specified as 0 or `adaptive`, adaptive method is used.
            If the value is specified as `sann`, sann algorithm is used.

        threshold : float, optional
            only used if ``cutoff=adaptive``. A threshold which is used as safe limit for calculation of cutoff.

        voroexp : int, optional
            only used if ``method=voronoi``. Power of the neighbor weight used to weight the contribution of each atom towards
            Steinhardt parameter values. Default 1.

        padding : double, optional
            only used if ``cutoff=adaptive`` or ``cutoff=number``. A safe padding value used after an adaptive cutoff is found. Default 1.2.

        nlimit : int, optional
            only used if ``cutoff=adaptive``. The number of particles to be considered for the calculation of adaptive cutoff.
            Default 6.
        
        cells : bool, optional
            If True, always use cell lists. Default None.

        nmax : int, optional
            only used if ``cutoff=number``. The number of closest neighbors to be found for each atom. Default 12
        

        Returns
        -------
        None

        Raises
        ------
        RuntimeWarning
            raised when `threshold` value is too low. A low threshold value will lead to 'sann' algorithm not converging
            when finding a neighbor. This function will try to automatically increase `threshold` and check again.

        RuntimeError
            raised when neighbor search was unsuccessful. This is due to a low `threshold` value.

        Notes
        -----
        This function calculates the neighbors of each particle. There are several ways to do this. A complete description of
        the methods can be `found here <https://pyscal.readthedocs.io/en/latest/nearestneighbormethods.html>`_.

        Method cutoff and specifying a cutoff radius uses the traditional approach being the one in which the neighbors of an atom
        are the ones that lie in the cutoff distance around it.

        In order to reduce time during the distance sorting during thefind_neighbors adaptive methods, pyscal sets an initial guess for a cutoff distance.
        This is calculated as,

        .. math:: r_{initial} = threshold * (simulation~box~volume/ number~of~particles)^{(1/3)}

        threshold is a safe multiplier used for the guess value and can be set using the `threshold` keyword.

        In Method cutoff, if ``cutoff='adaptive'``, an adaptive cutoff is found during runtime for each atom [1].
        Setting the cutoff radius to 0 also uses this algorithm. The cutoff for an atom i is found using,

        .. math:: r_c(i) = padding * ((1/nlimit) * \sum_{j=1}^{nlimit}(r_{ij}))

        padding is a safe multiplier to the cutoff distance that can be set through the keyword `padding`. `nlimit` keyword sets the
        limit for the top nlimit atoms to be taken into account to calculate the cutoff radius.

        In Method cutoff, if ``cutoff='sann'``, sann algorithm is used [2]. There are no parameters to tune sann algorithm.

        The second approach is using Voronoi polyhedra which also assigns a weight to each neighbor in the ratio of the face area between the two atoms.
        Higher powers of this weight can also be used [3]. The keyword `voroexp`
        can be used to set this weight.
        
        If method is `number`, instead of using a cutoff value for finding neighbors, a specified number of closest atoms are
        found. This number can be set through the argument `nmax`.

        If `cells` is None, cell lists are used if number of atoms are higher than 2500. If True, cell lists are always used.

        .. warning::

            Adaptive and number cutoff uses a padding over the intial guessed "neighbor distance". By default it is 2. In case
            of a warning that ``threshold`` is inadequate, this parameter should be further increased. High/low value
            of this parameter will correspond to the time taken for finding neighbors.

        References
        ----------
        .. [1] Stukowski, A, Model Simul Mater SC 20, 2012
        .. [2] van Meel, JA, Filion, L, Valeriani, C, Frenkel, D, J Chem Phys 234107, 2012
        .. [3] Haeberle, J, Sperl, M, Born, P, arxiv 2019

        """
        #first reset all neighbors
        self.reset_neighbors()
        self.filter = 0

        if threshold < 1:
            raise ValueError("value of threshold should be at least 1.00")

        if cells is None:
            cells = (self.natoms > 250)

        if method == 'cutoff':            
            if cutoff=='sann':    
                finished = False
                for i in range(1, 10):
                    finished = pc.get_all_neighbors_sann(self.atoms, 0.0, 
                        self.triclinic, self.rot, self.rotinv,
                        self.boxdims, threshold*i, cells)
                    if finished:
                        if i>1:
                            warnings.warn("found neighbors with higher threshold than default/user input")
                        break
                    warnings.warn("Could not find sann cutoff. trying with a higher threshold", RuntimeWarning)
                else:
                    raise RuntimeError("sann cutoff could not be converged. This is most likely, \
                        due to a low threshold value. Try increasing it.")
            
            elif cutoff=='adaptive' or cutoff==0:
                finished = pc.get_all_neighbors_adaptive(self.atoms, 0.0,
                    self.triclinic, self.rot, self.rotinv,
                    self.boxdims, threshold, nlimit, padding, cells)
                if not finished:
                    raise RuntimeError("Could not find adaptive cutoff")
            else:
                if cells:
                    pc.get_all_neighbors_cells(self.atoms, cutoff,
                        self.triclinic, self.rot, self.rotinv, self.boxdims)
                else:
                    pc.get_all_neighbors_normal(self.atoms, cutoff,
                        self.triclinic, self.rot, self.rotinv, self.boxdims)

        elif method == 'number':
            finished = pc.get_all_neighbors_bynumber(self.atoms, 0.0, 
                self.triclinic, self.rot, self.rotinv,
                self.boxdims, threshold, nmax, cells, assign_neighbor)
            if not finished:
                raise RuntimeError("Could not find enough neighbors - try increasing threshold")

        
        elif method == 'voronoi':
            clean_vertices = (cutoff>0)
            #CLEANING is TURNED OFF
            #if not clean_vertices:
                #copy the simulation cell
            backupbox = self._box.copy()
            if self.triclinic:
                if not self.ghosts_created:
                    atoms, box = self.repeat((1, 1, 1), ghost=True, scale_box=True, assign=False, return_atoms=True)
                    self._atoms = atoms
                    self = self.embed_in_cubic_box()
            pc.get_all_neighbors_voronoi(self.atoms, 0.0,
                self.triclinic, self.rot, self.rotinv,
                self.boxdims, voroexp)

            if self.triclinic:
                self._box = backupbox

            #now clean up
            #else:
            #    real_atomdict = {"positions":copy.copy(self.atoms.positions_for_all), 
            #     "ghost":copy.copy(self.atoms.positions_for_ghost)}
                #we need to call the method
                #this means alles good
            #    if self.actual_box is None:
            #        if self.triclinic:
            #            new_box = self.embed_in_cubic_box(inputbox=self._box, return_box=True)
            #            rot = np.array(new_box).T
            #            rotinv = np.linalg.inv(rot)
            #        else:
            #            new_box = self._box
            #            rot = [[0,0,0], [0,0,0], [0,0,0]]
            #            rotinv = [[0,0,0], [0,0,0], [0,0,0]]
            #    #ghosts are present
            #    else:
            #        if self.triclinic:
            #            new_box = self.embed_in_cubic_box(inputbox=self.actual_box, return_box=True)
            #            rot = np.array(new_box).T
            #            rotinv = np.linalg.inv(rot)
            #        else:
            #            new_box = self.actual_box
            #            rot = [[0,0,0], [0,0,0], [0,0,0]]
            #            rotinv = [[0,0,0], [0,0,0], [0,0,0]]

            #    boxdims = [0,0,0]
            #    boxdims[0] = np.sum(np.array(new_box[0])**2)**0.5
            #    boxdims[1] = np.sum(np.array(new_box[1])**2)**0.5
            #    boxdims[2] = np.sum(np.array(new_box[2])**2)**0.5
                
            #    pc.get_all_neighbors_voronoi(real_atomdict, 0.0,
            #        self.triclinic, rot, rotinv,
            #        boxdims, 1)                
                
            #    pc.clean_voronoi_vertices(real_atomdict, 
            #        self.atoms, 0.0,
            #        self.triclinic, rot, rotinv,
            #        boxdims, cutoff)

                #unique_vertices = []
                #for i in range(len(self.vertex_is_unique)):
                #    for j in range(len(self.vertex_is_unique[i])):
                #        if self.vertex_is_unique[i][j]:
                #            unique_vertices.append(self.vertex_positions[i][j])

                #self.atoms["vertex_positions_unique_skipcheck"] = unique_vertices

            #assign extra options
            mapdict = {}
            mapdict["voronoi"] = {}
            mapdict["voronoi"]["volume"] = "voronoi_volume"
            mapdict["voronoi"]["face"] = {}
            mapdict["voronoi"]["face"]["vertices"] = "face_vertices"
            mapdict["voronoi"]["face"]["perimeters"] = "face_perimeters"
            mapdict["voronoi"]["vertex"] = {}
            mapdict["voronoi"]["vertex"]["vectors"] = "vertex_vectors"
            mapdict["voronoi"]["vertex"]["numbers"] = "vertex_numbers"
            mapdict["voronoi"]["vertex"]["positions"] = "vertex_positions"
            #mapdict["voronoi"]["vertex"]["unique_positions"] = "vertex_positions_unique_nofilter"
            self.atoms._add_attribute(mapdict)

        self.neighbors_found = True

    def calculate_q(self, q, averaged=False, continuous_algorithm=False):
        """
        Find the Steinhardt parameter q_l for all atoms.

        Parameters
        ----------
        q : int or list of ints
            A list of all Steinhardt parameters to be found.

        averaged : bool, optional
            If True, return the averaged q values, default False
        
        continuous_algorithm: bool, optional
            See Notes for description.

        Returns
        -------
        q : list of floats
            calculated q values

        Notes
        -----
        Enables calculation of the Steinhardt parameters [1] q. The type of
        q values depend on the method used to calculate neighbors. See the description
        :func:`~pyscal.core.System.find_neighbors` for more details. 

        The option `continuous_algorithm` specifies which algorithm to use for calculations. If False, 
        an algorithm [3] is used. The C++ algorithm is faster is a large, consecutive number of q values (> 200)
        are to be calculated.

        This function creates three new attributes for this class: `qx`, `qx_real` and `qx_imag`,
        where `stands` for the q number.   

        References
        ----------
        .. [1] Steinhardt, PJ, Nelson, DR, Ronchetti, M. Phys Rev B 28, 1983
        .. [2] Lechner, W, Dellago, C, J Chem Phys, 2013
        """
        if isinstance(q, int):
            qq = [q]
        else:
            qq = q

        self._check_neighbors()

        if averaged:
            self._calculate_aq(qq)
            qvals = [self.atoms["avg_q%d"%x] for x in qq]
        else:    
            if continuous_algorithm:
                lm = max(qq)
                pc.calculate_q(self.atoms, lm)
            else:
                self._calculate_q(qq)
            qvals = [self.atoms["q%d"%x] for x in qq]
        return qvals

    def _calculate_q(self, qq):
        """
        Private method for calculation of qvals
        """
        for val in qq:
            pc.calculate_q_single(self.atoms, val)
  
        mapdict = {}
        mapdict["steinhardt"] = {}
        mapdict["steinhardt"]["generic"] = {}
        for val in qq:
            key1a = "q%d_norm"%val
            key1b = "q%d"%val
            key2 = "q%d_real"%val
            key3 = "q%d_imag"%val
            mapdict["steinhardt"]["generic"][key1a] = key1b
            mapdict["steinhardt"]["generic"][key2] = key2
            mapdict["steinhardt"]["generic"][key3] = key3
        self.atoms._add_attribute(mapdict)


    def _calculate_aq(self, qq):
        """
        Private method for calculation of avged qvals
        """

        todo_q = []
        for q in qq:
            keys = ["q%d"%q, "q%d_real"%q, "q%d_imag"%q]
            prod = []
            for key in keys:
                if key in self.atoms.keys():
                    prod.append(True)
                else:
                    prod.append(False)
            prod = np.prod(prod)
            if not prod:
                todo_q.append(q)

        _ = self._calculate_q(todo_q)

        #loop over atoms
        for val in qq:
            pc.calculate_aq_single(self.atoms, val)

        mapdict = {}
        mapdict["steinhardt"] = {}
        mapdict["steinhardt"]["average"] = {}
        for val in qq:
            key1a = "q%d_norm"%val
            key1b = "q%d"%val
            key2 = "q%d_real"%val
            key3 = "q%d_imag"%val
            mapdict["steinhardt"]["average"][key1a] = key1b
            mapdict["steinhardt"]["average"][key2] = key2
            mapdict["steinhardt"]["average"][key3] = key3
        self.atoms._add_attribute(mapdict)

    def calculate_disorder(self, averaged=False, q=6):
        """
        Calculate the disorder criteria for each atom
        
        Parameters
        ----------
        averaged : bool, optional
            If True, calculate the averaged disorder. Default False.
        q : int, optional
            The Steinhardt parameter value over which the bonds have to be calculated.
            Default 6.
        
        Returns
        -------
        None
        
        Notes
        -----
        Calculate the disorder criteria as introduced in [1]. The disorder criteria value for each atom is defined by,
        .. math::
            D_j = \\frac{1}{N_b^j} \sum_{i=1}^{N_b} [ S_{jj} + S_{kk} -2S_{jk}]
        where .. math:: S_{ij} = \sum_{m=-6}^6 q_{6m}(i) q_{6m}^*(i)
        
        Any q value other than six can also be used. This can be specified using the `q` argument.

        The keyword `averaged` is True, the disorder value is averaged over the atom and its neighbors. 
        For ordered systems, the value of disorder would be zero which would increase
        and reach one for disordered systems.

        This function creates two new attributes for this class: `disorder` and `avg_disorder`.
        
        References
        ----------
        .. [1] Kawasaki, T, Onuki, A, J. Chem. Phys. 135, 2011
        """
        #now routine for calculation of disorder

        keys = ["q%d_real"%q, "q%d_imag"%q]
        prod = []
        for key in keys:
            if key in self.atoms.keys():
                prod.append(True)
            else:
                prod.append(False)
        prod = np.prod(prod)
        if not prod:
            self.calculate_q(q)

        pc.calculate_disorder(self.atoms, q)

        mapdict = {}
        mapdict["steinhardt"] = {}
        mapdict["steinhardt"]["disorder"] = {}
        mapdict["steinhardt"]["disorder"]["norm"] = "disorder"

        if averaged:
            #average the disorder
            avg_arr = self.average_over_neighbors("disorder")
            self.atoms["avg_disorder"] = avg_arr
            mapdict["steinhardt"]["disorder"]["average"] = "avg_disorder"
        self.atoms._add_attribute(mapdict)


    def find_solids(self, bonds=0.5, threshold=0.5, avgthreshold=0.6, 
                          cluster=True, q=6, cutoff=0, right=True):
        """
        Distinguish solid and liquid atoms in the system.
        Parameters
        ----------
        bonds : int or float, optional
            Minimum number of solid bonds for an atom to be identified as
            a solid if the value is an integer. Minimum fraction of neighbors
            of an atom that should be solid for an atom to be solid if the
            value is float between 0-1. Default 0.5.
        
        threshold : double, optional
            Solid bond cutoff value. Default 0.5.
        
        avgthreshold : double, optional
            Value required for Averaged solid bond cutoff for an atom to be identified
            as solid. Default 0.6.
        
        cluster : bool, optional
            If True, cluster the solid atoms and return the number of atoms in the largest
            cluster.
        
        q : int, optional
            The Steinhardt parameter value over which the bonds have to be calculated.
            Default 6.
        
        cutoff : double, optional
            Separate value used for cluster classification. If not specified, cutoff used
            for finding neighbors is used.
        
        right: bool, optional
            If true, greater than comparison is to be used for finding solid particles. 
            default True.
        
        Returns
        -------
        solid : int
            Size of the largest solid cluster. Returned only if `cluster=True`.
        
        Notes
        -----
        The neighbors should be calculated before running this function.
        Check :func:`~pyscal.core.System.find_neighbors` method.
        
        `bonds` define the number of solid bonds of an atom to be identified as solid.
        Two particles are said to be 'bonded' if [1],
        .. math:: s_{ij} = \sum_{m=-6}^6 q_{6m}(i) q_{6m}^*(i) \geq threshold
        where `threshold` values is also an optional parameter.
        
        If the value of `bonds` is a fraction between 0 and 1, at least that much of an atom's neighbors
        should be solid for the atom to be solid.
        
        An additional parameter `avgthreshold` is an additional parameter to improve solid-liquid distinction.
        
        In addition to having a the specified number of `bonds`,
        
        .. math::  \langle s_{ij} \\rangle > avgthreshold
        
        also needs to be satisfied. In case another q value has to be used for calculation of S_ij, it can be
        set used the `q` attribute. In the above formulations, `>` comparison for `threshold` and `avgthreshold`
        can be changed to `<` by setting the keyword `right` to False.
        
        If `cluster` is True, a clustering is done for all solid particles. See :func:`~pyscal.csystem.find_clusters`
        for more details. 
        
        References
        ----------
        .. [1] Auer, S, Frenkel, D. Adv Polym Sci 173, 2005
        """
        #check if neighbors are found
        self._check_neighbors()

        if not isinstance(q, int):
            raise TypeError("q should be interger value")

        if not isinstance(threshold, (int, float)):
            raise TypeError("threshold should be a float value")
        else:
            if not ((threshold >= 0 ) and (threshold <= 1 )):
                raise ValueError("Value of threshold should be between 0 and 1")

        if not isinstance(avgthreshold, (int, float)):
            raise TypeError("avgthreshold should be a float value")
        else:
            if not ((avgthreshold >= 0 ) and (avgthreshold <= 1 )):
                raise ValueError("Value of avgthreshold should be between 0 and 1")

        #start identification routine
        #check the value of bonds and set criteria depending on that
        if isinstance(bonds, int):
            criteria = 0
        elif isinstance(bonds, float):
            if ((bonds>=0) and (bonds<=1.0)):
                criteria = 1
            else:
                raise TypeError("bonds if float should have value between 0-1")
        else:
             raise TypeError("bonds should be interger/float value")

        if right:
            compare_criteria = 0
        else:
            compare_criteria = 1

        self.calculate_q(q)

        #calculate solid neighs
        pc.calculate_bonds(self.atoms, q, 
            threshold, avgthreshold, bonds, 
            compare_criteria, criteria)

        mapdict = {}
        mapdict["steinhardt"] = {}
        mapdict["steinhardt"]["order"] = {}
        mapdict["steinhardt"]["order"]["bonds"] = "bonds"
        mapdict["steinhardt"]["order"]["sij"] = {}
        mapdict["steinhardt"]["order"]["sij"]["norm"] = "sij"
        mapdict["steinhardt"]["order"]["sij"]["average"] = "avg_sij"
        mapdict["steinhardt"]["order"]["sij"]["solid"] = "solid"
        self.atoms._add_attribute(mapdict)
        
        if cluster:
            lc = self.cluster_atoms(self.atoms.steinhardt.order.sij.solid, largest=True)
            return lc

    def find_largest_cluster(self):
        """
        Find largest cluster among given clusters
        
        Parameters
        ----------
        None

        Returns
        -------
        lc : int
            Size of the largest cluster.
        """
        if not "cluster" in self.atoms.keys():
            raise RuntimeError("cluster_atoms needs to be called first")

        clusterlist = [x for x in self.atoms["cluster"] if x != -1]
        xx, xxcounts = np.unique(clusterlist, return_counts=True)
        arg = np.argsort(xxcounts)[-1]
        largest_cluster_size = xxcounts[arg]
        largest_cluster_id = xx[arg]


        self.atoms["largest_cluster"] = [True if self.atoms["cluster"][x]==largest_cluster_id else False for x in range(len(self.atoms["cluster"]))]
        
        mapdict = {}
        mapdict["cluster"] = {}
        mapdict["cluster"]["largest"] = "largest_cluster"
        self.atoms._add_attribute(mapdict)

        return largest_cluster_size


    def cluster_atoms(self, condition, largest = True, cutoff=0):
        """
        Cluster atoms based on a property
        
        Parameters
        ----------
        condition : callable or atom property
            Either function which should take an :class:`~Atom` object, and give a True/False output
            or an attribute of atom class which has value or 1 or 0.
        
        largest : bool, optional
            If True returns the size of the largest cluster. Default False.
        
        cutoff : float, optional
            If specified, use this cutoff for calculation of clusters. By default uses the cutoff
            used for neighbor calculation.
        
        Returns
        -------
        lc : int
            Size of the largest cluster. Returned only if `largest` is True.
        
        Notes
        -----
        This function helps to cluster atoms based on a defined property. This property
        is defined by the user through the argument `condition` which is passed as a parameter.
        `condition` should be a boolean array the same length as number of atoms in the system.
        """
        
        self.apply_selection(condition=condition)
        pc.find_clusters(self.atoms, cutoff)

        mapdict = {}
        mapdict["cluster"] = {}
        mapdict["cluster"]["id"] = "cluster"
        self.atoms._add_attribute(mapdict)

        #done!
        lc = self.find_largest_cluster()
        #pcs.System.get_largest_cluster_atoms(self)
        self.remove_selection()
        if largest:
            return lc


    def calculate_rdf(self, rmin=0, rmax=5.00, bins=100):
        """
        Calculate the radial distribution function.
        
        Parameters
        ----------
        rmin : float, optional
            minimum value of the distance histogram. Default 0.0.
        
        rmax : float, optional
            maximum value of the distance histogram. Default 5.0

        bins : int
            number of bins in the histogram
                
        Returns
        -------
        rdf : array of ints
            Radial distribution function
        
        r : array of floats
            radius in distance units
        """
        self.find_neighbors(method="cutoff", cutoff=rmax)
        distances = list(itertools.chain(*self.atoms["neighbordist"]))

        hist, bin_edges = np.histogram(distances, bins=bins, 
            range=(rmin, rmax), density=True)
        edgewidth = np.abs(bin_edges[1]-bin_edges[0])
        hist = hist.astype(float)
        r = bin_edges[:-1]

        #get box density
        rho = self.natoms/self.volume
        shell_vols = (4./3.)*np.pi*((r+edgewidth)**3 - r**3)
        shell_rho = hist/shell_vols
        #now divide to get final value
        rdf = shell_rho/rho
        return rdf, r

    def calculate_angularcriteria(self):
        """
        Calculate the angular criteria for each atom
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        
        Notes
        -----
        Calculates the angular criteria for each atom as defined in [1]_. Angular criteria is
        useful for identification of diamond cubic structures. Angular criteria is defined by,
        
        .. math::
            A = \sum_{i=1}^6 (\cos(\\theta_i) + \\frac{1}{3})^2
        
        where cos(theta) is the angle size suspended by each pair of neighbors of the central
        atom. A will have a value close to 0 for structures if the angles are close to 109 degrees.
        The calculated A parameter for each atom can be accessed by system.angular
        
        References
        ----------
        .. [1] Uttormark, MJ, Thompson, MO, Clancy, P, Phys. Rev. B 47, 1993
        """
        self._check_neighbors()
        angulars = []

        for count, pos1 in enumerate(self.atoms["positions"]):
            
            dists = []
            distneighs = []
            distvectors = []

            for count2, neigh in enumerate(self.atoms["neighbors"][count]):
                pos2 = self.atoms["positions"][neigh]
                dist = self.atoms["neighbordist"][count][count2]
                vectors = self.atoms["diff"][count][count2]
                dists.append(dist)
                distneighs.append(neigh)
                distvectors.append(vectors)

            args = np.argsort(dists)
            #find top four
            topfourargs = np.array(args)[:4]

            combos = list(itertools.combinations(topfourargs, 2))
            costhetasum = 0

            for combo in combos:
                vec1 = distvectors[combo[0]]
                vec2 = distvectors[combo[1]]
                modvec1 = np.sqrt(np.sum([x**2 for x in vec1]))
                modvec2 = np.sqrt(np.sum([x**2 for x in vec2]))
                costheta = np.dot(vec1, vec2)/(modvec1*modvec2)
                costhetasum += (costheta +(1./3.))**2
            angulars.append(costhetasum)

        self.atoms["angular"] = angulars
        
        mapdict = {}
        mapdict["angular_parameters"] = {}
        mapdict["angular_parameters"]["diamond_angle"] = "angular"
        self.atoms._add_attribute(mapdict)

    def calculate_chiparams(self, angles=False):
        """
        Calculate the chi param vector for each atom
        
        Parameters
        ----------
        angles : bool, optional
            If True, return the list of cosines of all neighbor pairs
        
        Returns
        -------
        angles : array of floats
            list of all cosine values, returned only if `angles` is True.
        
        Notes
        -----
        This method tries to distinguish between crystal structures by finding the cosines of angles
        formed by an atom with its neighbors. These cosines are then historgrammed with bins
        `[-1.0, -0.945, -0.915, -0.755, -0.705, -0.195, 0.195, 0.245, 0.795, 1.0]` to find a vector for
        each atom that is indicative of its local coordination. Compared to chi parameters from chi_0 to
        chi_7 in the associated publication, the vector here is from chi_0 to chi_8. This is due to an additional
        chi parameter which measures the number of neighbors between cosines -0.705 to -0.195.
        Parameter `nlimit` specifies the number of nearest neighbors to be included in the analysis to find the cutoff.
        If parameter `angles` is true, an array of all cosine values is returned. The publication further provides
        combinations of chi parameters for structural identification which is not implemented here. The calculated
        chi params can be accessed using :attr:`~pyscal.catom.chiparams`.
        
        References
        ----------
        .. [1] Ackland, Jones, Phys. Rev. B 73, 2006
        """

        self._check_neighbors()

        bins = [-1.0, -0.945, -0.915, -0.755, -0.705, -0.195, 0.195, 0.245, 0.795, 1.0]
        chiparams = []
        cosines = []

        for count, pos in enumerate(self.atoms["positions"]):

            dists = self.atoms["neighbordist"][count]
            neighs = self.atoms["neighbors"][count]

            args = range(len(dists))
            combos = list(itertools.combinations(args, 2))
            costhetas = []
            
            for combo in combos:
                vec1 = self.atoms["diff"][count][combo[0]]
                vec2 = self.atoms["diff"][count][combo[1]]
                modvec1 = np.linalg.norm(vec1)
                modvec2 = np.linalg.norm(vec2)
                costheta = np.dot(vec1, vec2)/(modvec1*modvec2)
                #found costheta
                costhetas.append(costheta)


            #now add according to classification in paper
            chivector = np.histogram(costhetas, bins=bins)
            chiparams.append(chivector[0])
            if angles:
                cosines.append(costhetas)
        
        self.atoms["chiparams"] = chiparams
        
        mapdict = {}
        mapdict["angular_parameters"] = {}
        mapdict["angular_parameters"]["chi_params"] = "chiparams"
        
        if angles:
            self.atoms["cosines"] = cosines
            mapdict["angular_parameters"]["cosines"] = "cosines"

        self.atoms._add_attribute(mapdict)


import numpy as np
from pyscal.attributes import DocumentedKeywords
import spglib

def get_symmetry(system, angle_tolerance=-1.0,
        symprec=1e-5, hall_number=0):
    """
    Calculate space group number and other symmetry properties

    Parameters
    ----------
    system: pyscal System object
        the input system

    angle_tolerance: float, optional
        An experimental argument that controls angle tolerance between basis vectors. 
        Normally it is not recommended to use this argument. Default -1.0
        https://spglib.github.io/spglib/variable.html#angle-tolerance

    symprec: float, optional
        Distance tolerance in Cartesian coordinates to find crystal symmetry. 
        Default 1e-5
        https://spglib.github.io/spglib/variable.html#symprec

    hall_number: int, optional
        The argument to constrain the space-group-type search only for the Hall symbol corresponding to it. 
        Default 0
        https://spglib.github.io/spglib/dataset.html#hall-number

    Returns
    -------
    results: dict
        dict containing results. The keys are explained below:    
        international_space_group_number: 
            The space group type number defined in International Tables for Crystallography (ITA).
            https://it.iucr.org/Ac/contents/ 
        international_symbol:
            The (full) Hermann–Mauguin notation of space group type
            https://www.chemeurope.com/en/encyclopedia/Hermann-Mauguin_notation.html
        point_group:
            Symbol of the crystallographic point group in the Hermann–Mauguin notation 
            https://www.chemeurope.com/en/encyclopedia/Hermann-Mauguin_notation.html
    
    """
    res = spglib.get_symmetry_dataset((system.box,
        system.direct_coordinates, system.atoms.types))

    results = {}
    results['international_space_group_number'] = res["number"]
    results['international_symbol'] = res["international"]
    results['point_group'] = res["pointgroup"]
    return results



import numpy as np
import gzip
from ase import Atom, Atoms
import gzip
import io
import os

#new function to wrap over mdtraj objects
def read_snap(mdobject, check_triclinic=False):
    """
    Function to read from an MDTraj atoms objects

    Parameters
    ----------
    mdobject : MDTraj Atoms object
        name of the MDTraj atoms object

    triclinic : bool, optional
        True if the configuration is triclinic

    """
    #We have to process atoms and atomic objects from ase
    #Known issues lammps -dump modified format
    #first get box
    a = np.array(mdobject.unitcell_vectors[0][0])
    b = np.array(mdobject.unitcell_vectors[0][1])
    c = np.array(mdobject.unitcell_vectors[0][2])

    box = np.array([a, b, c])

    #box and box dims are set. Now handle atoms
    chems = np.array([atom.name for atom in mdobject.topology.atoms])
    atomsymbols = np.unique(chems)
    atomtypes = np.array(range(1, len(atomsymbols)+1))
    typedict = dict(zip(atomsymbols, atomtypes))

    #now start parsing atoms
    positions = mdobject.xyz[0]
    ids = [count+1 for count in range(len(positions))]
    species = [chems[count] for count in range(len(positions))]

    atoms = {}
    atoms['positions'] = positions
    atoms['ids'] = ids
    atoms['species'] = species
    atoms['ghost'] = [False for x in range(len(types))]
    return atoms, box

def write_snap(**kwargs):
    raise NotImplementedError("write method for mdtraj is not implemented")

def split_snaps(**kwargs):
    raise NotImplementedError("split method for mdtraj is not implemented")

def convert_snap(**kwargs):
    raise NotImplementedError("convert method for mdtraj is not implemented")



import numpy as np
import gzip
from ase import Atom, Atoms
import gzip
import io
import os

#new function to wrap over ase objects
def read_snap(aseobject, check_triclinic=False):
    """
    Function to read from a ASE atoms objects

    Parameters
    ----------
    aseobject : ASE Atoms object
        name of the ASE atoms object

    triclinic : bool, optional
        True if the configuration is triclinic

    """
    #We have to process atoms and atomic objects from ase
    #Known issues lammps -dump modified format
    #first get box
    a = np.array(aseobject.cell[0])
    b = np.array(aseobject.cell[1])
    c = np.array(aseobject.cell[2])

    box = np.array([a, b, c])

    #box and box dims are set. Now handle atoms
    chems = np.array(aseobject.get_chemical_symbols())
    atomsymbols = np.unique(aseobject.get_chemical_symbols())
    atomtypes = np.array(range(1, len(atomsymbols)+1))
    typedict = dict(zip(atomsymbols, atomtypes))

    #now start parsing atoms
    positions = aseobject.positions
    ids = []
    types = []
    species = []

    for count, position in enumerate(positions):
        ids.append(count+1)
        types.append(typedict[chems[count]])
        species.append(chems[count])

    atoms = {}
    atoms['positions'] = positions
    atoms['ids'] = ids
    atoms['types'] = types
    atoms['species'] = species
    atoms['ghost'] = [False for x in range(len(types))]

    return atoms, box

def write_snap(**kwargs):
    raise NotImplementedError("write method for mdtraj is not implemented")

def split_snaps(**kwargs):
    raise NotImplementedError("split method for mdtraj is not implemented")
    
def convert_snap(sys, species=None):
    """
    Convert a given pyscal structure to ase object

    Parameters
    ----------
    sys : System object
        the system object to be converted

    species : list of str
        a list of species in the system

    Returns
    -------
    aseobject: ASE atoms object

    Notes
    -----
    ASE needs the species of atoms. If a property called `species`
    exist in :attr:`~pyscal.catom.Atom.custom`, this value is used for
    species. However if the value is not present, the keyword `species`
    is required. This should contain a mapping between :attr:`~pyscal.catom.Atom.type`
    and species name. For example, if `species` is `['Au', 'Ge']`, all atoms
    of type 1 are assigned as Au and those of type 2 are assigned as Ge.
    Note that ase is required to run this method.

    """
    #we only do a local import of ASE, this is not super nice
    #we can change this later depending on if ASE is to be treated
    #as a full dependency
    
    #get element strings
    if 'species' not in sys.atoms.keys():
        sys.atoms["species"] = [None for x in range(sys.atoms.ntotal)]

    if sys.atoms.species[0] is None:
        if species is None:
            raise ValueError("Species was not known! To convert to ase, species need to be provided using the species keyword")
        #otherwise we know the species
        types = sys.atoms.types
        unique_types = np.unique(types)
        if not (len(unique_types) == len(species)):
            raise ValueError("Length of species and number of types found in system are different. Maybe you specified \"Au\" instead of [\"Au\"]")
        #now assign the species to custom
        atomspecies = []        
        for cc, typ in enumerate(types):
            atomspecies.append(species[int(typ-1)])
    else:
        atomspecies = sys.atoms.species
      
    cell = sys.box
    pbc = [1, 1, 1]

    #create ASE Atoms and assign everything
    aseobject = Atoms()
    aseobject.cell = cell
    aseobject.pbc = pbc
    
    #thats everything pretty much
    #now create ase Atom
    for count, pos in enumerate(sys.atoms.positions):
        aseatom = Atom(atomspecies[count], pos)
        aseobject.append(aseatom)
    
    #done
    return aseobject



import numpy as np
from pyscal.attributes import DocumentedKeywords

def repeat(system, repetitions, ghost = False,
    scale_box = True, atoms = None, return_atoms = False, return_box=False):
    """
    Repeat the given system

    Parameters
    ----------
    system: pyscal System object
        the input system to be repeated

    repetitions: tuple of ints
        number of times the system is to be rotated in each direction

    ghost: bool, optional
        if True, make the new atoms ghost, default False

    scale_box: bool, optional
        if True, scale the simulation box, default True

    atoms: None, optional
        if provided use the given atoms, and not the atoms from the system

    return_atoms: bool, optional
        if True, return atoms instead of adding them to the system. Default False
    
    Returns
    -------
    system: pyscal System object
        the system with repetitions. Only returned if `return_atoms` is False.
    
    atoms: Atoms object
        only returned if `return_atoms` is True.


    """
    box = system.box        
    system.actual_box = box.copy()

    if atoms is None:
        atoms = system.atoms

    newatoms = []
    idstart = len(atoms) + 1

    x1 = -repetitions[0]
    x2 = repetitions[0]+1
    y1 = -repetitions[1]
    y2 = repetitions[1]+1
    z1 = -repetitions[2]
    z2 = repetitions[2]+1
    xs = 2*repetitions[0] + 1
    ys = 2*repetitions[1] + 1
    zs = 2*repetitions[2] + 1
    
    datadict = {key:[] for key in atoms.keys()}
    del datadict['positions']
    del datadict['ids']
    del datadict['head']
    del datadict['ghost']
    positions = []
    ids = []
    head = []
    ghosts = []

    for i in range(x1, x2):
        for j in range(y1, y2):
            for k in range(z1, z2):
                if (i==j==k==0):
                    continue
                for count, pos in enumerate(atoms['positions']):
                    #we should create ghost images for only real atoms
                    if not atoms["ghost"][count]:
                        pos = (pos + i*np.array(box[0]) + j*np.array(box[1]) + k*np.array(box[2]))
                        positions.append(list(pos))
                        ids.append(idstart)
                        head.append(count)
                        ghosts.append(ghost)
                        idstart += 1
                        for key in datadict.keys():
                            datadict[key].append(atoms[key][count])

    if scale_box:
        box[0] = xs*np.array(box[0])
        box[1] = ys*np.array(box[1])
        box[2] = zs*np.array(box[2])
    if ghost:
        system.ghosts_created = True

    atoms['positions'].extend(positions)
    atoms['ids'].extend(ids)
    atoms['ghost'].extend(ghosts)
    atoms['head'].extend(head)
    for key in datadict.keys():
        atoms[key].extend(datadict[key])

    if return_atoms:
        return atoms, box

    else:
        system.box = box
        system.atoms = atoms
        return system


def embed_in_cubic_box(system, input_box=None, 
    return_box=False):
    """
    Embedded the triclinic box in a cubic box
    
    Parameters
    ----------
    system : pyscal System
        input system

    input_box: optional
        if specified, this box is used instead of the
        system simulation box

    return_box: bool, optional
        if specified, return the box instead of assigning
        to the system. Default False

    Returns
    -------
    system: 
        system object with the modified box
        only returned if `return_box` is False

    box:
        the cubic simulation box
        only returned if `return_box` is True
    """
    if input_box is None:
        box = system._box
        backupbox = box.copy()
    else:
        box = input_box
        backupbox = input_box.copy

    a = np.array(box[0])
    b = np.array(box[1])
    c = np.array(box[2])

    cosa = np.dot(b, c)/(np.linalg.norm(b)*np.linalg.norm(c))
    cosb = np.dot(c, a)/(np.linalg.norm(c)*np.linalg.norm(a))
    cosc = np.dot(a, b)/(np.linalg.norm(a)*np.linalg.norm(b))

    lx = np.linalg.norm(a)
    xy = np.linalg.norm(b)*cosc
    xz = np.linalg.norm(c)*cosb
    ly = np.sqrt(np.linalg.norm(b)*np.linalg.norm(b) - xy*xy)
    yz = (np.linalg.norm(b)*np.linalg.norm(c)*cosa - xy*xz)/ly
    lz = np.sqrt(np.linalg.norm(c)*np.linalg.norm(c) - xz*xz - yz*yz)

    xlo = ylo = zlo = 0
    xhi = lx
    yhi = ly
    zhi = lz

    xlo_bound = xlo + min(0.0,xy,xz,xy+xz)
    xhi_bound = xhi + max(0.0,xy,xz,xy+xz)
    ylo_bound = ylo + min(0.0,yz)
    yhi_bound = yhi + max(0.0,yz)
    zlo_bound = zlo
    zhi_bound = zhi

    newbox = np.array([[xhi_bound-xlo_bound, 0, 0], [0, yhi_bound-ylo_bound, 0], [0, 0, zhi_bound-zlo_bound]])
    
    if not return_box:
        system.newbox = newbox
        system.box = newbox
        system.box_backup = backupbox
        system.actual_box = None
        return system
    else:
        return newbox



from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import setup, find_packages


with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    name='pyscal3',
    version='3.0.0.dev.6',
    author='Sarath Menon',
    author_email='sarath.menon@pyscal.org',
    description='Python library written in C++ for calculation of local atomic structural environment',
    long_description=readme,
    long_description_content_type = "text/markdown",
    # tell setuptools to look for any packages under 'src'
    packages=find_packages('src'),
    # tell setuptools that all packages will be under the 'src' directory
    # and nowhere else
    package_dir={'':'src'},
    # add an extension module named 'python_cpp_example' to the package
    #headers=["src/pyscal/atom.h", "src/pyscal/system.h", "lib/voro++/voro++.hh"],
    #headers=["src/pyscal/atom.h", "src/pyscal/modsystem.h"],
    headers=["src/pyscal/modsystem.h"],
    ext_modules=[
        #Pybind11Extension(
        #    "pyscal.catom",
        #    ["src/pyscal/atom.cpp", "src/pyscal/atom_binding.cpp"],
        #    language='c++',
        #    include_dirs=['lib/voro++']
        #),
        Pybind11Extension(
            "pyscal.csystem",
            ["src/pyscal/neighbor.cpp", "src/pyscal/sh.cpp", 
            "src/pyscal/solids.cpp", "src/pyscal/voronoi.cpp", 
            "src/pyscal/modsystem_binding.cpp", "lib/voro++/voro++.cc"],
            language='c++',
            include_dirs=['lib/voro++'],
            extra_compile_args=['-O3'],
        ),

        #Pybind11Extension(
        #    "pyscal.csystem",
        #    ["src/pyscal/system.cpp", "src/pyscal/system_binding.cpp", "src/pyscal/atom.cpp", "lib/voro++/voro++.cc"],
        #    language='c++',
        #    include_dirs=['lib/voro++']
        #),
    ],
    # add custom build_ext command
    cmdclass={"build_ext": build_ext},
    zip_safe=False,
    download_url = 'https://anaconda.org/conda-forge/pyscal',
    url = 'https://pyscal.org',
    install_requires=['pybind11', 'numpy', 'ase'],
    classifiers=[
        'Programming Language :: Python :: 3'
    ],
    include_package_data=True,
    #package_data={'': ['*.yaml']},
)



import pytest
import os,sys,inspect
import numpy as np
import pyscal.core as pc
from pyscal.crystal_structures import Structure

def test_angular():
    sys = Structure().lattice.diamond(repetitions = [4, 4, 4])
    sys.find_neighbors(method = 'cutoff', cutoff=0)
    sys.calculate_angularcriteria()

    assert np.round(np.mean(np.array(sys.atoms.angular_parameters.diamond_angle)), decimals=2) == 0.00
